import os
import json
from typing import Optional


def doccano2nerblackbox(_input_file: str, _output_file: str, verbose: bool = False):
    if verbose:
        print(f"> read input_file = {_input_file}")
    with open(_input_file, "r") as f:
        input_lines = [json.loads(line) for line in f]

    output_lines = list()
    for input_line in input_lines:
        output_line = {
            "text": input_line["text"],
            "tags": [
                {
                    "char_start": label[0],
                    "char_end": label[1],
                    "token": input_line["text"][label[0] : label[1]],
                    "tag": label[2],
                }
                for label in input_line["label"]
            ],
        }
        output_lines.append(output_line)

    if verbose:
        print(f"> write output_file = {_output_file}")

    os.makedirs("/".join(_output_file.split("/")[:-1]), exist_ok=True)
    with open(_output_file, "w") as f:
        for line in output_lines:
            f.write(json.dumps(line, ensure_ascii=False) + "\n")

    return _output_file


def nerblackbox2doccano(
    _input_file: str, _max_lines: Optional[int] = None, verbose: bool = False
) -> str:

    if verbose:
        print(f"> read input_file = {_input_file}")
    with open(_input_file, "r") as f:
        input_lines = [json.loads(line) for line in f]

    if _max_lines is not None:
        input_lines = input_lines[:_max_lines]

    output_lines = list()
    for input_line in input_lines:
        output_line = {
            "text": input_line["text"],
            "label": [
                [int(tag["char_start"]), int(tag["char_end"]), tag["tag"]]
                for tag in input_line["tags"]
            ],
        }
        output_lines.append(output_line)

    output_file = _input_file.replace(".jsonl", "_DOCCANO.jsonl")
    if verbose:
        print(f"> write output_file = {output_file}")
    with open(output_file, "w") as f:
        for line in output_lines:
            f.write(json.dumps(line, ensure_ascii=False) + "\n")

    return output_file



GENERAL = {
    "experiment_name": "str",
    "from_config": "bool",
    "run_name": "str",
    "run_name_nr": "str",
    "device": "str",
    "fp16": "bool",
    "experiment_run_name_nr": "str",
}

DATASET = {
    "dataset_name": "str",
    "annotation_scheme": "str",
    "prune_ratio_train": "float",
    "prune_ratio_val": "float",
    "prune_ratio_test": "float",
    "train_on_val": "bool",
    "train_on_test": "bool",
}
MODEL = {
    "pretrained_model_name": "str",
}
SETTINGS = {
    "uncased": "bool",
    "checkpoints": "bool",
    "logging_level": "str",
    "multiple_runs": "int",
    "seed": "int",
}
PARAMS = {**DATASET, **MODEL, **SETTINGS}

HPARAMS = {
    "batch_size": "int",
    "max_seq_length": "int",
    "max_epochs": "int",
    "early_stopping": "bool",
    "monitor": "str",
    "min_delta": "float",
    "patience": "int",
    "mode": "str",
    "lr_max": "float",
    "lr_schedule": "str",
    "lr_warmup_epochs": "int",
    "lr_cooldown_epochs": "int",
    "lr_cooldown_restarts": "bool",
    "lr_num_cycles": "int",
}
LOG_DIRS = {
    "mlflow": "str",
    "tensorboard": "str",
    "checkpoints": "str",
    "log_file": "str",
    "mlflow_file": "str",
}



from typing import List
import pandas as pd


class Annotation:
    """
    Attributes:
        classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        scheme:                                  e.g. "plain" or "bio"
    """

    def __init__(self, classes: List[str]):
        """
        Args:
            classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        """
        self.classes: List[str] = classes
        self.scheme: str = ""
        self._process()

    def _process(self):
        """
        changed attributes:
            scheme: [str], e.g. "plain" or "bio"
            classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        """
        self._get_scheme()  # attr: scheme
        if self.scheme == "bio":
            self._ensure_completeness_in_case_of_bio_tags()  # attr: classes
        elif self.scheme == "bilou":
            self._ensure_completeness_in_case_of_bilou_tags()  # attr: classes
        self._sort_classes()  # attr: classes

    def _get_scheme(self) -> None:
        """
        derive annotation scheme from classes

        changed attributes:
            scheme: [str], e.g. "plain" or "bio"
        """
        if any(["-" in tag for tag in self.classes]):
            if any(["L-" in tag for tag in self.classes]) or any(
                ["U-" in tag for tag in self.classes]
            ):
                self.scheme = "bilou"
            else:
                self.scheme = "bio"
        else:
            self.scheme = "plain"

    def _ensure_completeness_in_case_of_bio_tags(self) -> None:
        """
        in case of BIO-tags: make sure that there is an "I-*" tag for every "B-*" tag

        changed attributes:
            classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        """
        b_tags = [tag for tag in self.classes if tag.startswith("B")]
        for b_tag in b_tags:
            i_tag = b_tag.replace("B-", "I-")
            if i_tag not in self.classes:
                self.classes.append(i_tag)

    def _ensure_completeness_in_case_of_bilou_tags(self) -> None:
        """
        in case of BILOU-tags: make sure that there is an "I-*", "L-*", "U-" tag for every "B-*" tag

        changed attributes:
            classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        """
        b_tags = [tag for tag in self.classes if tag.startswith("B")]
        for b_tag in b_tags:
            for x in ["I-", "L-", "U-"]:
                x_tag = b_tag.replace("B-", x)
                if x_tag not in self.classes:
                    self.classes.append(x_tag)

    def _sort_classes(self) -> None:
        """
        sort classes. in case of BIO-tags, all B-* tags come first.

        changed attributes:
            classes: tag classes present in dataset, e.g. ["PER", "ORG"] or ["B-person", "B-time", "I-person"]
        """
        self.classes = ["O"] + sorted([elem for elem in self.classes if elem != "O"])

    def change_scheme(self, new_scheme: str) -> "Annotation":
        if self.scheme == new_scheme:
            return self
        elif new_scheme == "plain":
            classes_bio_without_o = [elem for elem in self.classes if elem != "O"]
            classes_plain = ["O"] + list(
                pd.Series(classes_bio_without_o)
                .map(lambda x: x.split("-")[-1])
                .drop_duplicates()
            )

            annotation_plain = Annotation(classes_plain)
            return annotation_plain
        elif new_scheme in ["bio", "bilou"]:
            classes_plain_without_o = list(
                set([elem.split("-")[-1] for elem in self.classes if elem != "O"])
            )
            classes_bio_bilou = ["O"]
            for elem in classes_plain_without_o:
                if new_scheme == "bio":
                    classes_bio_bilou += [f"B-{elem}", f"I-{elem}"]
                else:  # bilou
                    classes_bio_bilou += [
                        f"B-{elem}",
                        f"I-{elem}",
                        f"L-{elem}",
                        f"U-{elem}",
                    ]

            annotation_bio_bilou = Annotation(classes_bio_bilou)
            return annotation_bio_bilou
        else:
            raise Exception(f"ERROR! new_scheme = {new_scheme} is not implemented.")



import pytest
from typing import Union, Dict, Optional, List

from pkg_resources import resource_filename
import os

from nerblackbox.modules.experiment_config.experiment_config import ExperimentConfig

os.environ["DATA_DIR"] = resource_filename("nerblackbox", f"tests/test_data")


class TestExperimentConfig:

    experiment_config_default = ExperimentConfig(experiment_name="default")
    experiment_config = ExperimentConfig(experiment_name="test_experiment")

    # 1 ################################################################################################################
    def test_no_config_file(self):
        with pytest.raises(Exception):
            _ = ExperimentConfig(experiment_name="test_experiment_that_does_not_exist")

    # 2 ################################################################################################################
    @pytest.mark.parametrize(
        "default, config_dict",
        [
            (
                False,
                {
                    "params": {
                        "dataset_name": "swedish_ner_corpus",
                        "annotation_scheme": "plain",
                        "prune_ratio_train": 0.01,
                        "prune_ratio_val": 0.01,
                        "prune_ratio_test": 0.01,
                        "train_on_val": True,
                        "train_on_test": True,
                        "pretrained_model_name": "KB/bert-base-swedish-cased",
                        "multiple_runs": 1,
                        "uncased": False,
                    },
                    "hparams": {
                        "max_epochs": 3,
                        "lr_max": 2e-5,
                        "lr_schedule": "hybrid",
                        "early_stopping": False,
                    },
                    "runA": {
                        "lr_cooldown_epochs": 1,
                    },
                },
            ),
            (
                True,
                {
                    "params": {
                        "prune_ratio_train": 1.00,
                        "prune_ratio_val": 1.00,
                        "prune_ratio_test": 1.00,
                        "train_on_val": False,
                        "train_on_test": False,
                        "checkpoints": True,
                        "logging_level": "info",
                        "multiple_runs": 1,
                        "seed": 42,
                    },
                    "hparams": {
                        "batch_size": 16,
                        "max_seq_length": 64,
                        "max_epochs": 50,
                        "early_stopping": True,
                        "monitor": "val_loss",
                        "mode": "min",
                        "min_delta": 0.0,
                        "patience": 0,
                        "lr_warmup_epochs": 2,
                        "lr_cooldown_epochs": 3,
                        "lr_cooldown_restarts": False,
                        "lr_num_cycles": 4,
                    },
                },
            ),
        ],
    )
    def test_get_config(self, default: bool, config_dict: Dict[str, Dict[str, str]]):
        if default:
            test_config_dict = self.experiment_config_default.config
        else:
            test_config_dict = self.experiment_config.config
        assert sorted(list(test_config_dict.keys())) == sorted(
            list(config_dict.keys())
        ), (
            f"ERROR! test_config_dict.keys() = {sorted(list(test_config_dict.keys()))} does not equal "
            f"config_dict.keys() = {sorted(list(config_dict.keys()))}"
        )

        keys = list(test_config_dict.keys())
        for key in keys:
            assert sorted(test_config_dict[key].items()) == sorted(
                config_dict[key].items()
            ), (
                f"ERROR! sorted(test_config_dict[{key}].items()) = {sorted(test_config_dict[key].items())} "
                f"!= {sorted(config_dict[key].items())}"
            )

    # 3 ################################################################################################################
    @pytest.mark.parametrize(
        "input_key, input_value, converted_input, error",
        [
            (
                "lr_schedule",
                "constant",
                "constant",
                False,
            ),
            (
                "prune_ratio_train",
                "0.01",
                0.01,
                False,
            ),
            (
                "checkpoints",
                "False",
                False,
                False,
            ),
            (
                "multiple_runs",
                "5",
                5,
                False,
            ),
            (
                "xyz",
                "",
                "",
                True,
            ),
        ],
    )
    def test_convert(
        self,
        input_key: str,
        input_value: str,
        converted_input: Union[str, int, float, bool],
        error: bool,
    ):
        if error:
            with pytest.raises(Exception):
                _ = self.experiment_config._convert(input_key, input_value)
        else:
            test_converted_input = self.experiment_config._convert(
                input_key, input_value
            )
            assert test_converted_input == converted_input, (
                f"ERROR! test_converted_input = {test_converted_input} != {converted_input} "
                f"for input_key = {input_key} and input_value = {input_value}"
            )



from os.path import isfile, join
import glob
from typing import Optional, Any, Dict, Union, Tuple
from pkg_resources import Requirement
from pkg_resources import resource_filename
import mlflow
from nerblackbox.api.utils import Utils
from nerblackbox.api.store import Store
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.parameters import DATASET, MODEL, SETTINGS, HPARAMS
from nerblackbox.modules.experiment_config.preset import get_preset
from nerblackbox.modules.experiment_results import ExperimentResults


class Experiment:
    def __init__(
        self,
        experiment_name: str,
        from_config: bool = False,
        model: Optional[str] = None,
        dataset: Optional[str] = None,
        from_preset: Optional[str] = "adaptive",
        pytest: bool = False,
        **kwargs_optional: Any,
    ):
        self.experiment_name = experiment_name
        self.from_preset = from_preset

        self.from_config: bool
        self.kwargs: Dict[str, str]
        self.hparams: Optional[Dict[str, str]]
        self.results: Optional[ExperimentResults]

        if not pytest:
            experiment_exists, experiment_results = Store.get_experiment_results_single(
                experiment_name
            )
            if experiment_exists:
                self.from_config = True
                self.results = experiment_results
                print(f"> experiment = {experiment_name} found, results loaded.")
            else:
                self.from_config = from_config
                self.kwargs, self.hparams = self._parse_arguments(
                    model, dataset, self.from_preset, **kwargs_optional
                )
                self._checks()  # self.hparams, self.from_preset, self.from_config
                self.results = None
                print(
                    f"> experiment = {experiment_name} not found, create new experiment."
                )

    def show_config(self) -> None:
        r"""
        print experiment config

        Used Attr:
            experiment_name: [str] e.g. 'exp0' or 'all'
        """
        from nerblackbox.modules.utils.env_variable import env_variable

        if self.experiment_name != "all":
            path_experiment_config = join(
                env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
            )
            if isfile(path_experiment_config):
                with open(path_experiment_config, "r") as file:
                    lines = file.read()

                print(f"> experiment_config = {path_experiment_config}")
                print()
                print(lines)
            else:
                print(f"> experiment_config = {path_experiment_config} does not exist.")
        else:
            experiment_configs = glob.glob(
                join(env_variable("DIR_EXPERIMENT_CONFIGS"), "*.ini")
            )
            experiment_configs = [
                elem.split("/")[-1].strip(".ini") for elem in experiment_configs
            ]
            experiment_configs = [
                elem for elem in experiment_configs if elem != "default"
            ]
            for experiment_config in experiment_configs:
                print(experiment_config)

    def run(self):
        r"""run a single experiment.

        Note:

        - from_config == True -> experiment config file is used, no other optional arguments will be used

        - from_config == False -> experiment config file is created dynamically, optional arguments will be used

            - model and dataset are mandatory.

            - All other arguments relate to hyperparameters and are optional.
              They are determined using the following hierarchy:

              1) optional argument

              2) from_preset (adaptive, original, stable),
                 which specifies e.g. the hyperparameters "max_epochs", "early_stopping", "lr_schedule"

              3) default experiment configuration
        """
        _parameters = {
            "experiment_name": self.experiment_name,
            "from_config": int(self.from_config),
            "run_name": self.kwargs["run_name"],
            "device": self.kwargs["device"],
            "fp16": self.kwargs["fp16"],
        }

        if self.from_config is False:
            assert self.hparams is not None, f"ERROR! self.hparams is None."
            self._write_config_file(hparams=self.hparams)

        mlflow.projects.run(
            uri=resource_filename(Requirement.parse("nerblackbox"), "nerblackbox"),
            entry_point="run_experiment",
            experiment_name=self.experiment_name,
            parameters=_parameters,
            env_manager="local",
        )

        experiment_exists, self.results = Store.get_experiment_results_single(
            self.experiment_name
        )
        assert (
            experiment_exists
        ), f"ERROR! experiment = {self.experiment_name} does not exist."
        print("### single runs ###")
        print(self.results.single_runs.T)
        print()
        print("### average runs ###")
        print(self.results.average_runs.T)

    def get_result(
        self,
        metric: str = "f1",
        level: str = "entity",
        label: str = "micro",
        phase: str = "test",
        average: bool = False,
    ) -> Optional[str]:
        r"""

        Args:
            metric: "f1", "precision", "recall"
            level: "entity" or "token"
            label: "micro", "macro", "PER", ..
            phase: "val" or "test"
            average: if True, return average result of all runs. if False, return result of best run.

        Returns:
            result: e.g. "0.9011 +- 0.0023" (average = True) or "0.9045" (average = False)
        """
        if self.results is None:
            print(f"ATTENTION! no results found")
            return None
        else:
            key = f"{phase.upper()}_{level[:3].upper()}_{metric.upper()}"
            base_quantity = (
                self.results.best_average_run
                if average
                else self.results.best_single_run
            )
            assert isinstance(
                base_quantity, dict
            ), f"ERROR! type(base_quantity) = {type(base_quantity)} should be dict."

            if key in base_quantity:
                if isinstance(
                    base_quantity[key], str
                ):  # average = True,  e.g. "0.9011 +- 0.0023"
                    return base_quantity[key]
                elif isinstance(
                    base_quantity[key], float
                ):  # average = False, e.g. 0.9045..
                    return f"{base_quantity[key]:.4f}"
                else:
                    raise Exception(
                        f"ERROR! found result of unexpected type = {type(base_quantity[key])}"
                    )
            else:
                print(f"ATTENTION! no results found")
                return None

    ####################################################################################################################
    # HELPER METHODS
    ####################################################################################################################
    def _write_config_file(self, hparams: Dict[str, str]) -> None:
        r"""
        write config file based on self.hparams
        """
        # assert that config file does not exist
        config_path = join(
            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
        )
        assert (
            isfile(config_path) is False
        ), f"ERROR! experiment config file {config_path} already exists!"

        # write config file: helper functions
        def _write(_str: str):
            f.write(_str + "\n")

        def _write_key_value(_key: str):
            assert (
                hparams is not None
            ), f"ERROR! self.hparams is None - _write_key_value() failed."
            if _key in hparams.keys():
                f.write(f"{_key} = {hparams[_key]}\n")

        # write config file
        with open(config_path, "w") as f:
            _write("[dataset]")
            for key in DATASET.keys():
                _write_key_value(key)

            _write("\n[model]")
            for key in MODEL.keys():
                _write_key_value(key)

            _write("\n[settings]")
            for key in SETTINGS.keys():
                _write_key_value(key)

            _write("\n[hparams]")
            for key in HPARAMS.keys():
                _write_key_value(key)

            _write("\n[runA]")

    def _parse_arguments(
        self,
        model: Optional[str] = None,
        dataset: Optional[str] = None,
        from_preset: Optional[str] = "adaptive",
        **kwargs_optional: Any,
    ) -> Tuple[Dict[str, str], Optional[Dict[str, str]]]:
        kwargs = Utils.process_kwargs_optional(kwargs_optional)

        if model is not None:
            kwargs["pretrained_model_name"] = model
        if dataset is not None:
            kwargs["dataset_name"] = dataset

        _hparams = Utils.extract_hparams(kwargs)
        kwargs["hparams"] = self._process_hparams(_hparams, from_preset)
        kwargs["run_name"] = kwargs["run_name"] if "run_name" in kwargs else ""
        kwargs["device"] = kwargs["device"] if "device" in kwargs else "gpu"
        kwargs["fp16"] = int(kwargs["fp16"]) if "fp16" in kwargs else 0

        if not self.from_config:
            kwargs["from_preset"] = from_preset
            hparams = kwargs.pop("hparams")

            # get rid of keys in kwargs that are present in hparams
            for key in hparams.keys():
                if key in kwargs.keys():
                    kwargs.pop(key)
        else:
            hparams = None

        return kwargs, hparams

    @staticmethod
    def _process_hparams(
        hparams: Optional[Dict[str, Union[str, int, bool]]], from_preset: Optional[str]
    ) -> Optional[Dict[str, Union[str, int, bool]]]:
        """
        Args:
            hparams:         [dict], e.g. {'multiple_runs': '2'} with hparams to use            [HIERARCHY:  I]
            from_preset:     [str], e.g. 'adaptive' get experiment params & hparams from preset [HIERARCHY: II]

        Returns:
            _hparams:        [dict], e.g. {'multiple_runs': '2'} with hparams to use
        """
        _hparams = get_preset(from_preset)

        if _hparams is None:
            _hparams = hparams
        elif hparams is not None:
            _hparams.update(**hparams)

        return _hparams

    def _checks(self):
        """
        checks that
        - either static or dynamic approach is used, not a mixture
        - if static, that experiment config exists
        - if dynamic, that both model and dataset are specified
        """
        # assert STATIC or DYNAMIC
        assert (self.hparams is None and self.from_config is True) or (
            self.hparams is not None and self.from_config is False
        ), (
            f"ERROR! Need to specify "
            f"EITHER hparams (currently {self.hparams}) "
            f"with or without from_preset (currently {self.from_preset}) "
            f"OR from_config (currently {self.from_config})."
        )

        if self.from_config:
            # STATIC - assert that experiment config exists
            path_experiment_config = join(
                env_variable("DIR_EXPERIMENT_CONFIGS"),
                f"{self.experiment_name}.ini",
            )
            if not isfile(path_experiment_config):
                self._exit_gracefully(
                    f"experiment_config = {path_experiment_config} does not exist."
                )
        else:
            # DYNAMIC - assert that model & dataset are specified
            assert self.hparams is not None, f"ERROR! self.hparams is None."
            for field in ["pretrained_model_name", "dataset_name"]:
                if field not in self.hparams.keys():
                    field_displayed = (
                        "model" if field == "pretrained_model_name" else "dataset"
                    )
                    self._exit_gracefully(
                        f"{field_displayed} is not specified but mandatory if dynamic arguments are used."
                    )

    @staticmethod
    def _exit_gracefully(message: str) -> None:
        print(message)
        print("stopped.")
        exit(0)



import pytest
from typing import Optional, Dict, Union

from nerblackbox.api.experiment import Experiment
import os
from os.path import join, abspath, dirname
from pkg_resources import resource_filename

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR
TRACKING_URI = resource_filename("nerblackbox", f"data/results/mlruns")


class TestApiExperiment:

    experiment = Experiment("xyz", pytest=True)

    # 1 ################################################################################################################
    @pytest.mark.parametrize(
        "hparams, from_preset, hparams_processed",
        [
            (
                None,
                None,
                None,
            ),
            (
                {"multiple_runs": "2"},
                None,
                {"multiple_runs": "2"},
            ),
            (
                {"multiple_runs": "2"},
                "adaptive",
                {
                    "multiple_runs": "2",
                    "max_epochs": 250,
                    "early_stopping": True,
                    "lr_schedule": "constant",
                },
            ),
        ],
    )
    def test_process_hparams(
        self,
        hparams: Optional[Dict[str, Union[str, int, bool]]],
        from_preset: Optional[str],
        hparams_processed: Optional[Dict[str, str]],
    ):
        test_hparams_processed = self.experiment._process_hparams(hparams, from_preset)
        assert (
            test_hparams_processed == hparams_processed
        ), f"ERROR! test_hparams_processed = {test_hparams_processed} != {hparams_processed}"



import pytest
import pandas as pd
from nerblackbox.modules.datasets.formatter.huggingface_datasets_formatter import (
    HuggingfaceDatasetsFormatter,
)
from nerblackbox.modules.datasets.formatter.base_formatter import SENTENCES_ROWS
from pkg_resources import resource_filename
from typing import List, Tuple, Optional, Dict

import os
from os.path import abspath, dirname, join

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR


class TestHuggingfaceDatasetsFormatter:

    formatter_pretokenized = HuggingfaceDatasetsFormatter(
        "conll2003"
    )  # use any ner_dataset that works
    formatter_pretokenized_subset = HuggingfaceDatasetsFormatter(
        "wikiann", "ace"
    )  # use any ner_dataset that works
    formatter_unpretokenized = HuggingfaceDatasetsFormatter(
        "ehealth_kd"
    )  # use any ner_dataset that works

    ####################################################################################################################
    # check_existence()
    ####################################################################################################################
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, existence",
        [
            (["ehealth_kd", "", True]),
            (["conll2003", "", True]),
            (["conll2003", "this_subset_does_not_exist", False]),
            (["wikiann", "ace", True]),
            (["wikiann", "this_subset_does_not_exist", False]),
            (["swedish_ner_corpus", "", True]),
            (["sent_comp", "", True]),
            (["this_dataset_does_not_exist", "", False]),
        ],
    )
    def test_check_existence(
        self, ner_dataset: str, ner_dataset_subset: str, existence: bool
    ):
        test_existence, _ = HuggingfaceDatasetsFormatter.check_existence(
            ner_dataset, ner_dataset_subset
        )
        assert (
            test_existence == existence
        ), f"ERROR! test_existence = {test_existence} != {existence} for ner_dataset = {ner_dataset}"

    ####################################################################################################################
    # check_compatibility()
    ####################################################################################################################
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, compatibility",
        [
            (["ehealth_kd", "", True]),
            (["conll2003", "", True]),
            (["wikiann", "ace", True]),
            (
                [
                    "swedish_ner_corpus",
                    "",
                    False,  # only train & test on huggingface datasets
                ]
            ),
            (
                ["sent_comp", "", False]
            ),  # only train & validation on huggingface datasets
        ],
    )
    def test_check_compatibility(
        self, ner_dataset: str, ner_dataset_subset: str, compatibility: bool
    ):
        test_compability, _ = HuggingfaceDatasetsFormatter.check_compatibility(
            ner_dataset, ner_dataset_subset
        )
        assert (
            test_compability == compatibility
        ), f"ERROR! test_compatibility = {test_compability} != {compatibility} for ner_dataset = {ner_dataset}"

    ####################################################################################################################
    # check_implementation()
    ####################################################################################################################
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, implementation",
        [
            (["ehealth_kd", "", True]),  # ['entities'][0]['ent_label'].names
            (["conll2003", "", True]),  # ['ner_tags'].feature.names
            (["wikiann", "ace", True]),  # ['ner_tags'].feature.names
            (["swedish_ner_corpus", "", True]),  # ['ner_tags'].feature.names
            (["sent_comp", "", False]),
        ],
    )
    def test_check_implementation(
        self, ner_dataset: str, ner_dataset_subset: str, implementation: bool
    ):
        test_implementation, _ = HuggingfaceDatasetsFormatter.check_implementation(
            ner_dataset, ner_dataset_subset
        )
        assert (
            test_implementation == implementation
        ), f"ERROR! test_implementation = {test_implementation} != {implementation} for ner_dataset = {ner_dataset}"

    ####################################################################################################################
    # get_info()
    ####################################################################################################################
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, info",
        [
            (
                [
                    "ehealth_kd",
                    "",
                    (
                        True,  # ['entities'][0]['ent_label'].names
                        ["Concept", "Action", "Predicate", "Reference"],
                        False,
                        {
                            "text": "sentence",
                            "tags": "entities",
                            "mapping": {
                                "ent_text": "token",
                                "ent_label": "tag",
                                "start_character": "char_start",
                                "end_character": "char_end",
                            },
                        },
                    ),
                ]
            ),
            (
                [
                    "conll2003",
                    "",
                    (
                        True,  # ['ner_tags'].feature.names
                        [
                            "O",
                            "B-PER",
                            "I-PER",
                            "B-ORG",
                            "I-ORG",
                            "B-LOC",
                            "I-LOC",
                            "B-MISC",
                            "I-MISC",
                        ],
                        True,
                        {"text": "tokens", "tags": "ner_tags", "mapping": None},
                    ),
                ]
            ),
            (
                [
                    "wikiann",
                    "ace",
                    (
                        True,  # ['ner_tags'].feature.names
                        [
                            "O",
                            "B-PER",
                            "I-PER",
                            "B-ORG",
                            "I-ORG",
                            "B-LOC",
                            "I-LOC",
                        ],
                        True,
                        {"text": "tokens", "tags": "ner_tags", "mapping": None},
                    ),
                ]
            ),
            (
                [
                    "swedish_ner_corpus",
                    "",
                    (
                        True,  # ['ner_tags'].feature.names
                        ["0", "LOC", "MISC", "ORG", "PER"],
                        True,
                        {"text": "tokens", "tags": "ner_tags", "mapping": None},
                    ),
                ]
            ),
            (
                [
                    "sent_comp",
                    "",
                    (
                        False,
                        None,
                        None,
                        None,
                    ),
                ]
            ),
        ],
    )
    def test_get_info(
        self,
        ner_dataset: str,
        ner_dataset_subset: str,
        info: Tuple[
            bool, Optional[List[str]], Optional[bool], Optional[Dict[str, str]]
        ],
    ):
        print(ner_dataset)
        test_info = HuggingfaceDatasetsFormatter.get_infos(
            ner_dataset, ner_dataset_subset
        )
        print(test_info)
        for k in range(len(test_info)):
            assert (
                test_info[k] == info[k]
            ), f"ERROR! test_info[{k}] = {test_info[k]} != {info[k]} for ner_dataset = {ner_dataset}"

    ####################################################################################################################
    # get_ner_tag_list()
    ####################################################################################################################
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, ner_tag_list",
        [
            (["conll2003", "", ["LOC", "MISC", "ORG", "PER"]]),
            (["wikiann", "ace", ["LOC", "ORG", "PER"]]),
            (["ehealth_kd", "", ["Action", "Concept", "Predicate", "Reference"]]),
        ],
    )
    def test_get_ner_tag_list(
        self, ner_dataset: str, ner_dataset_subset: str, ner_tag_list: List[str]
    ):
        formatter = HuggingfaceDatasetsFormatter(ner_dataset, ner_dataset_subset)
        assert (
            formatter.ner_tag_list == ner_tag_list
        ), f"ERROR! test_ner_tag_list = {formatter.ner_tag_list} != {ner_tag_list}"

    ####################################################################################################################
    # ---
    ####################################################################################################################
    @pytest.mark.parametrize(
        "pretokenized, sentences_rows_iob2",
        [
            (
                True,
                [
                    [["Peter", "B-PER"], ["Blackburn", "I-PER"]],
                    [
                        ["EU", "B-ORG"],
                        ["rejects", "O"],
                        ["German", "B-MISC"],
                        ["call", "O"],
                        ["to", "O"],
                        ["boycott", "O"],
                        ["British", "B-MISC"],
                        ["lamb", "O"],
                        [".", "O"],
                    ],
                ],
            ),
            (
                False,
                [
                    {
                        "text": "La mayoría de las arritmias son el resultado de problemas en el sistema eléctrico del corazón.",
                        "tags": [
                            {
                                "token": "mayoría",
                                "tag": "Predicate",
                                "char_start": 95,
                                "char_end": 102,
                            },
                            {
                                "token": "arritmias",
                                "tag": "Concept",
                                "char_start": 110,
                                "char_end": 119,
                            },
                            {
                                "token": "problemas",
                                "tag": "Action",
                                "char_start": 140,
                                "char_end": 149,
                            },
                            {
                                "token": "sistema eléctrico",
                                "tag": "Concept",
                                "char_start": 156,
                                "char_end": 173,
                            },
                            {
                                "token": "corazón",
                                "tag": "Concept",
                                "char_start": 178,
                                "char_end": 185,
                            },
                        ],
                    },
                    {
                        "text": "En la leucemia linfocítica crónica, hay demasiados linfocitos, un tipo de glóbulos blancos.",
                        "tags": [
                            {
                                "token": "leucemia linfocítica crónica",
                                "tag": "Concept",
                                "char_start": 6,
                                "char_end": 34,
                            },
                            {
                                "token": "linfocitos",
                                "tag": "Concept",
                                "char_start": 51,
                                "char_end": 61,
                            },
                            {
                                "token": "glóbulos blancos",
                                "tag": "Concept",
                                "char_start": 74,
                                "char_end": 90,
                            },
                        ],
                    },
                ],
            ),
        ],
    )
    def test_format_data(self, pretokenized: bool, sentences_rows_iob2: SENTENCES_ROWS):

        formatter = (
            self.formatter_pretokenized
            if pretokenized
            else self.formatter_unpretokenized
        )

        formatter.get_data(verbose=False)
        if pretokenized:
            formatter.sentences_rows_pretokenized = {
                phase: formatter.sentences_rows_pretokenized[phase][:2]
                for phase in formatter.PHASES
            }
        else:
            formatter.sentences_rows_unpretokenized = {
                phase: formatter.sentences_rows_unpretokenized[phase][:2]
                for phase in formatter.PHASES
            }

        test_sentences_rows_iob2 = formatter.format_data(shuffle=True, write_csv=False)
        assert (
            test_sentences_rows_iob2 == sentences_rows_iob2
        ), f"ERROR! test_sentences_rows_iob2 = {test_sentences_rows_iob2} != {sentences_rows_iob2}"

    @pytest.mark.parametrize(
        "df_train, df_val, df_test",
        [
            (
                [
                    pd.DataFrame(
                        data=[
                            ["O O", "Mening 1"],
                            ["PER O", "Mening 2"],
                            ["O PER", "Mening 3"],
                            ["O PER", "Mening 4"],
                        ]
                    ),
                    pd.DataFrame(data=[["O O", "Mening 5"], ["PER O", "Mening 6"]]),
                    pd.DataFrame(data=[["O O", "Mening 7"], ["PER O", "Mening 8"]]),
                ]
            ),
        ],
    )
    def test_resplit_data(
        self, df_train: pd.DataFrame, df_val: pd.DataFrame, df_test: pd.DataFrame
    ):
        self.formatter_pretokenized.dataset_path = resource_filename(
            "nerblackbox", f"tests/test_data/formatted_data"
        )
        test_dfs = self.formatter_pretokenized.resplit_data(write_csv=False)
        assert (
            test_dfs is not None
        ), f"ERROR! resplit_data() returned None unexpectedly."
        test_df_train, test_df_val, test_df_test = test_dfs
        pd.testing.assert_frame_equal(
            test_df_train, df_train
        ), f"ERROR! test_resplit_data did not pass test for phase = train"
        pd.testing.assert_frame_equal(
            test_df_val, df_val
        ), f"ERROR! test_resplit_data did not pass test for phase = val"
        pd.testing.assert_frame_equal(
            test_df_test, df_test
        ), f"ERROR! test_resplit_data did not pass test for phase = test"



from dataclasses import dataclass
from dataclasses import asdict
from typing import List, Tuple, Callable

import numpy as np
from sklearn.metrics import accuracy_score as accuracy_sklearn
from sklearn.metrics import precision_score as precision_sklearn
from sklearn.metrics import recall_score as recall_sklearn
from sklearn.metrics import precision_recall_fscore_support as prf_sklearn
from sklearn.exceptions import UndefinedMetricWarning

import warnings

from seqeval.metrics import precision_score as precision_seqeval
from seqeval.metrics import recall_score as recall_seqeval
from seqeval.metrics import f1_score as f1_seqeval
from seqeval.scheme import IOB2, BILOU

from nerblackbox.modules.ner_training.annotation_tags.tags import Tags


class NerMetrics:
    """
    On the token  level, the tags are evaluated in the given annotation scheme (e.g. plain, BIO)
    On the entity level, the tags are evaluated in the BIO scheme (after converting if needed)
    """

    def __init__(
        self,
        true_flat,
        pred_flat,
        level,
        scheme,
        classes=None,
        class_index=None,
        verbose=False,
    ):
        """
        :param true_flat:   [np array] of shape [batch_size * seq_length]
        :param pred_flat:   [np array] of shape [batch_size * seq_length]
        :param level:       [str] 'token' or 'entity'
        :param scheme:      [str] e.g. 'plain', 'bio'
        :param classes:     [optional, list] of [str] labels to take into account for metrics -> if level = 'token'
        :param class_index: [optional, int]            index to take into account for metrics -> if level = 'entity'
        :param verbose:     [optional, bool] if True, show verbose output
        """
        self.true_flat = true_flat  # token -> plain. entity -> plain, bio, bilou
        self.pred_flat = pred_flat  # token -> plain. entity -> plain, bio, bilou
        self.scheme = scheme  # token -> plain. entity -> plain, bio, bilou
        self.classes = classes
        self.class_index = class_index
        self.level = level
        self.verbose = verbose

        if self.scheme == "bilou":
            self.scheme_entity = "bilou"
            self.scheme_entity_seqeval = BILOU
        else:  # plain, bio
            self.scheme_entity = "bio"
            self.scheme_entity_seqeval = IOB2

        self.results = Results()
        self.failure_value = -1

        assert self.level in [
            "token",
            "entity",
        ], f"ERROR! level = {self.level} unknown."
        if self.level == "entity":
            self.true_flat_bio: List[str] = Tags(self.true_flat,).convert_scheme(
                source_scheme=self.scheme, target_scheme=self.scheme_entity
            )  # entity -> bio, bilou

            self.pred_flat_bio: List[str] = Tags(self.pred_flat).convert_scheme(
                source_scheme=self.scheme, target_scheme=self.scheme_entity
            )  # entity -> bio, bilou

            # ASR
            self.pred_flat_bio_corrected: List[str]
            self.pred_flat_bio_corrected, self.results.asr_abidance = Tags(
                self.pred_flat_bio
            ).restore_annotation_scheme_consistency(
                scheme=self.scheme_entity
            )  # entity -> bio, bilou

    def results_as_dict(self):
        return asdict(self.results)

    def compute(self, _metrics):
        """
        computes selected metrics
        ----------------------------------------------------------
        :param _metrics: [list] of [str], e.g. ['acc, 'precision']
        :return: -
        """
        warnings.filterwarnings("error")

        if "acc" in _metrics:
            self.accuracy()

        if "precision" in _metrics or "recall" in _metrics or "f1" in _metrics:
            self._compute_well_defined_classes()
            if "precision" in _metrics or "f1" in _metrics:
                self.precision()
            if "recall" in _metrics or "f1" in _metrics:
                self.recall()
            if "f1" in _metrics:
                self.f1_score()

        if (
            "asr_abidance" in _metrics
            or "asr_precision" in _metrics
            or "asr_recall" in _metrics
            or "asr_f1" in _metrics
        ):
            self.compute_asr_results()

        warnings.resetwarnings()

    def accuracy(self):
        """
        computes accuracy of predictions (_np_logits) w.r.t. ground truth (_np_label_ids)
        ---------------------------------------------------------------------------------
        :return: acc [np float]
        """
        self.results.acc = accuracy_sklearn(
            self.true_flat, self.pred_flat, normalize=True
        )

    def precision(self):
        """
        computes precision (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)

        Returns:
            precision_micro [np array] for all examples
            precision_macro [np array] for each class, then averaged
        """
        if self.level == "token":
            self.results.precision_micro = self._token_evaluation(
                evaluation_function=precision_sklearn, average="micro"
            )
            self.results.precision_macro = self._token_evaluation(
                evaluation_function=precision_sklearn, average="macro"
            )

        elif self.level == "entity":
            self.results.precision_micro = self._entity_evaluation_micro(
                evaluation_function=precision_seqeval
            )
            self.results.precision_macro = self._entity_evaluation_macro(
                evaluation_function=precision_seqeval,
            )

    def recall(self):
        """
        computes recall (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)

        Returns:
            recall_micro [np array] for all examples
            recall_macro [np array] for each class, then averaged
        """
        if self.level == "token":
            self.results.recall_micro = self._token_evaluation(
                evaluation_function=recall_sklearn, average="micro"
            )
            self.results.recall_macro = self._token_evaluation(
                evaluation_function=recall_sklearn, average="macro"
            )
        elif self.level == "entity":
            self.results.recall_micro = self._entity_evaluation_micro(
                evaluation_function=recall_seqeval
            )
            self.results.recall_macro = self._entity_evaluation_macro(
                evaluation_function=recall_seqeval
            )

    def f1_score(self):
        """
        computes f1 score (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)

        Returns:
            f1_score_micro [np array] for all examples
            f1_score_macro [np array] for each class, then averaged
        """
        if self.level == "token":
            self.results.f1_micro = self._token_evaluation(
                evaluation_function=prf_sklearn, average="micro"
            )
            self.results.f1_macro = self._token_evaluation(
                evaluation_function=prf_sklearn, average="macro"
            )
        elif self.level == "entity":
            self.results.f1_micro, self.results.f1_macro = self._entity_evaluation_f1(
                evaluation_function=f1_seqeval,
            )

    def compute_asr_results(self):
        """
        computes
        - self.results.asr_precision_micro
        - self.results.asr_recall_micro
        - self.results.asr_f1_micro
        """

        def _entity_evaluation_micro_asr(evaluation_function: Callable) -> float:
            """helper function"""
            try:
                metric = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio_corrected],  # corrected !!!
                    average="micro",
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                metric = self.failure_value
            return metric

        self.results.asr_precision_micro = _entity_evaluation_micro_asr(
            evaluation_function=precision_seqeval
        )
        self.results.asr_recall_micro = _entity_evaluation_micro_asr(
            evaluation_function=recall_seqeval
        )
        self.results.asr_f1_micro = _entity_evaluation_micro_asr(
            evaluation_function=f1_seqeval
        )

    def _token_evaluation(self, evaluation_function: Callable, average: str) -> float:
        """
        compute precision/recall/f1 on token level

        Args:
            evaluation_function: precision_sklearn, recall_sklearn, prf_sklearn
            average: 'micro' or 'macro'

        Returns:
            metric: precision/recall on token level, 'micro' or 'macro' averaged
        """
        assert evaluation_function in [
            precision_sklearn,
            recall_sklearn,
            prf_sklearn,
        ], f"evaluation function = {evaluation_function} unknown / not allowed."
        assert average in ["micro", "macro"], f"average = {average} unknown."

        if self.classes is None or len(self.classes) > 1:  # "all" / "fil"
            if evaluation_function != prf_sklearn:
                metric = evaluation_function(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.classes,
                    average=average,
                    zero_division=0,
                )
            else:
                _, _, metric, _ = prf_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.classes,
                    average=average,
                    zero_division=0,
                )
        else:
            try:
                if evaluation_function != prf_sklearn:
                    metric = evaluation_function(
                        self.true_flat,
                        self.pred_flat,
                        labels=self.classes,
                        average=average,
                        zero_division="warn",
                    )
                else:
                    _, _, metric, _ = prf_sklearn(
                        self.true_flat,
                        self.pred_flat,
                        labels=self.classes,
                        average=average,
                        warn_for=("precision", "recall", "f-score"),
                        zero_division="warn",
                    )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                metric = self.failure_value

        return metric

    def _entity_evaluation_micro(self, evaluation_function: Callable) -> float:
        """
        compute precision/recall micro average on entity level

        Args:
            evaluation_function: precision_seqeval, recall_seqeval

        Returns:
            metric: precision/recall on entity level, 'macro' averaged
        """
        assert evaluation_function in [
            precision_seqeval,
            recall_seqeval,
        ], f"evaluation function = {evaluation_function} unknown / not allowed."

        if self.class_index is None:  # "fil"
            try:
                metric = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio],
                    average="micro",
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                metric = self.failure_value
        else:  # "ind"
            try:
                metric = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio],
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                    average=None,
                    zero_division="warn",
                )[self.class_index]
            except UndefinedMetricWarning:
                try:
                    metric = evaluation_function(
                        [self.true_flat_bio],
                        [self.pred_flat_bio],
                        mode="strict",
                        scheme=self.scheme_entity_seqeval,
                        average=None,
                        zero_division=0,
                    )[self.class_index]
                except IndexError:
                    metric = self.failure_value

                if metric == 0:
                    metric = evaluation_function(
                        [self.true_flat_bio],
                        [self.pred_flat_bio],
                        mode="strict",
                        scheme=self.scheme_entity_seqeval,
                        average=None,
                        zero_division=1,
                    )[self.class_index]
                    if metric == 1:
                        metric = self.failure_value
            except IndexError:
                metric = self.failure_value

        return metric

    def _compute_well_defined_classes(self) -> None:
        """
        Created Attributes:
            results.classindices_macro: list of indices of well-defined classes in terms of precision, recall, f1
            results.numberofclasses_macro: number of well-defined classes in terms of precision, recall, f1
        """

        def _get_index_list(
            evaluation_function: Callable, true_array, pred_array, scheme_seqeval=None
        ):
            kwargs = (
                {"mode": "strict", "scheme": scheme_seqeval}
                if scheme_seqeval is not None
                else {}
            )
            try:
                metric_list = evaluation_function(
                    true_array,
                    pred_array,
                    average=None,
                    zero_division="warn",
                    **kwargs,
                )
                index_list = [i for i in range(len(metric_list))]
            except UndefinedMetricWarning:
                metric_list_all = evaluation_function(
                    true_array,
                    pred_array,
                    average=None,
                    zero_division=0,
                    **kwargs,
                )

                index_list = list()
                for index, metric_elem in enumerate(metric_list_all):
                    if metric_elem != 0:
                        index_list.append(index)
                    else:
                        metric_elem_alt = evaluation_function(
                            true_array,
                            pred_array,
                            average=None,
                            zero_division=1,
                            **kwargs,
                        )[index]
                        if metric_elem_alt != 1:
                            index_list.append(index)

            return index_list

        if self.level == "token":
            index_list_precision = _get_index_list(
                evaluation_function=precision_sklearn,
                true_array=self.true_flat,
                pred_array=self.pred_flat,
            )
            index_list_recall = _get_index_list(
                evaluation_function=recall_sklearn,
                true_array=self.true_flat,
                pred_array=self.pred_flat,
            )
        else:
            index_list_precision = _get_index_list(
                evaluation_function=precision_seqeval,
                true_array=[self.true_flat_bio],
                pred_array=[self.pred_flat_bio],
                scheme_seqeval=self.scheme_entity_seqeval,
            )
            index_list_recall = _get_index_list(
                evaluation_function=recall_seqeval,
                true_array=[self.true_flat_bio],
                pred_array=[self.pred_flat_bio],
                scheme_seqeval=self.scheme_entity_seqeval,
            )

        self.results.classindices_macro = tuple(
            [index for index in index_list_precision if index in index_list_recall]
        )
        if self.level == "token":
            self.results.numberofclasses_macro = (
                len(self.results.classindices_macro) - 1
            )  # disregard "O" label
        else:
            self.results.numberofclasses_macro = len(self.results.classindices_macro)

    def _entity_evaluation_macro(
        self,
        evaluation_function: Callable,
    ) -> float:
        """
        compute precision/recall macro average on entity level

        Args:
            evaluation_function: precision_seqeval, recall_seqeval

        Returns:
            metric: precision/recall on entity level, 'macro' averaged on well-defined classes
        """
        assert evaluation_function in [
            precision_seqeval,
            recall_seqeval,
        ], f"evaluation function = {evaluation_function} unknown / not allowed."

        metric = evaluation_function(
            [self.true_flat_bio],
            [self.pred_flat_bio],
            mode="strict",
            scheme=self.scheme_entity_seqeval,
            average="macro",
            zero_division=0,
        )
        return metric

    def _entity_evaluation_f1(
        self, evaluation_function: Callable
    ) -> Tuple[float, float]:
        """
        compute f1 micro or macro average on entity level

        Args:
            evaluation_function: f1_seqeval

        Returns:
            f1_micro: f1 on entity level, 'micro' averaged
            f1_macro: f1 on entity level, 'macro' averaged on well-defined classes
        """
        assert evaluation_function in [
            f1_seqeval
        ], f"evaluation function = {evaluation_function} unknown / not allowed."

        # ensure that precision and recall have been called:
        # self.precision()
        # self.recall()

        # f1_micro
        if (
            self.results.precision_micro == self.failure_value
            or self.results.recall_micro == self.failure_value
        ):
            f1_micro = self.failure_value
        else:
            if self.class_index is None:  # "fil"
                f1_micro = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio],
                    average="micro",
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                )
            else:  # "ind"
                f1_micro = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio],
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                    average=None,
                    zero_division="warn",
                )[self.class_index]

        # f1_macro
        if (
            self.results.precision_macro == self.failure_value
            or self.results.recall_macro == self.failure_value
        ):
            f1_macro = self.failure_value
        else:
            if self.class_index is None:  # "fil"
                metric_list = evaluation_function(
                    [self.true_flat_bio],
                    [self.pred_flat_bio],
                    mode="strict",
                    scheme=self.scheme_entity_seqeval,
                    average=None,
                )
                f1_macro = np.average(metric_list)
            else:  # "ind"
                f1_macro = self.failure_value

        return f1_micro, f1_macro


@dataclass
class Results:
    acc: float = -1
    precision_micro: float = -1
    precision_macro: float = -1
    recall_micro: float = -1
    recall_macro: float = -1
    f1_micro: float = -1
    f1_macro: float = -1
    classindices_macro: Tuple[float, ...] = ()
    numberofclasses_macro: float = -1
    asr_abidance: float = -1
    asr_precision_micro: float = -1
    asr_recall_micro: float = -1
    asr_f1_micro: float = -1



import subprocess
from typing import Dict, List, Optional, Tuple
import pandas as pd

from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS,
    SENTENCES_ROWS_PRETOKENIZED,
)
from nerblackbox.modules.utils.env_variable import env_variable


class SwedishNerCorpusFormatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "swedish_ner_corpus"
        ner_tag_list = ["PER", "ORG", "LOC", "MISC", "PRG", "ORG*"]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]
        """
        bash_cmds = [
            f'git clone https://github.com/klintan/swedish-ner-corpus.git {env_variable("DIR_DATASETS")}/_swedish_ner_corpus',
            f'mv {env_variable("DIR_DATASETS")}/_swedish_ner_corpus/* {env_variable("DIR_DATASETS")}/swedish_ner_corpus',
            f'rm -rf {env_variable("DIR_DATASETS")}/_swedish_ner_corpus',
        ]

        for bash_cmd in bash_cmds:
            if verbose:
                print(bash_cmd)

            try:
                subprocess.run(bash_cmd, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                print(e)

    def create_ner_tag_mapping(self) -> Dict[str, str]:
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return {
            "ORG*": "ORG",
            "PRG": "O",
        }

    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            sentences_rows: only if write_csv = False
        """
        for phase in ["train", "test"]:
            sentences_rows = self._read_original_file(phase)
            if shuffle:
                sentences_rows = self._shuffle_dataset(phase, sentences_rows)

            if write_csv:  # pragma: no cover
                self._write_formatted_csv(phase, sentences_rows)
            else:
                return sentences_rows  # returns train!
        return None

    def set_original_file_paths(self) -> None:
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        self.file_name = {
            "train": "train_corpus.txt",
            "val": "val_corpus.txt",
            "test": "test_corpus.txt",
        }

    def _parse_row(self, _row: str) -> List[str]:
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        return _row.strip().split()

    def _format_original_file(self, _row_list: List[str]) -> Optional[List[str]]:
        """
        III: format data

        Args:
            _row_list: e.g. ["test", "PER", "X", "B"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        if not len(_row_list) in [2]:
            print(
                f"ATTENTION! row_list = {_row_list} should consist of 2 parts! -> treat as empty row"
            )
            return None

        _row_list_formatted = _row_list
        return _row_list_formatted

    def resplit_data(
        self, val_fraction: float, write_csv: bool = True
    ) -> Optional[Tuple[pd.DataFrame, ...]]:
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            df_train: only if write_csv = False
            df_val:   only if write_csv = False
            df_test:  only if write_csv = False
        """
        # train -> train, val
        df_train_val = self._read_formatted_csvs(["train"])
        df_train, df_val = self._split_off_validation_set(df_train_val, val_fraction)

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])

        if write_csv:  # pragma: no cover
            self._write_final_csv("train", df_train)
            self._write_final_csv("val", df_val)
            self._write_final_csv("test", df_test)
            return None
        else:
            return df_train, df_val, df_test



import pytest
from typing import Dict

import pandas as pd
from nerblackbox.modules.datasets.formatter.swedish_ner_corpus_formatter import (
    SwedishNerCorpusFormatter,
)

from pkg_resources import resource_filename
import os
from os.path import abspath, dirname, join

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR


class TestAnalyzer:

    analyzer = SwedishNerCorpusFormatter().analyzer
    analyzer.dataset_path = resource_filename(
        "nerblackbox", f"tests/test_data/final_data"
    )

    index = ["O", "PER", "ORG", "LOC", "MISC", "PRG", "ORG*"]
    columns = ["tags", "tags/sentence", "tags relative w/ 0", "tags relative w/o 0"]

    @pytest.mark.parametrize(
        "phase, num_sentences, stats_aggregated",
        [
            (
                "train",
                4,
                pd.DataFrame(data=[5, 3, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
            ),
            (
                "val",
                2,
                pd.DataFrame(data=[3, 1, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
            ),
            (
                "test",
                2,
                pd.DataFrame(data=[3, 1, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
            ),
        ],
    )
    def test_read_final_csv(
        self, phase: str, num_sentences: int, stats_aggregated: pd.DataFrame
    ):
        test_num_sentences, test_stats_aggregated = self.analyzer._read_final_csv(phase)
        assert (
            test_num_sentences == num_sentences
        ), f"ERROR! test_num_sentences = {test_num_sentences} not equal to num_sentences = {num_sentences}"
        pd.testing.assert_frame_equal(
            test_stats_aggregated, stats_aggregated
        ), f"ERROR! test_read_final_csv did not pass test for phase = {phase}"

    @pytest.mark.parametrize(
        "stats_aggregated_phase, num_tokens",
        [
            (
                pd.DataFrame(data=[3, 1, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
                4,
            ),
            (
                pd.DataFrame(data=[5, 3, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
                8,
            ),
        ],
    )
    def test_get_tokens(self, stats_aggregated_phase: pd.DataFrame, num_tokens):
        test_num_tokens = self.analyzer._get_num_tokens(stats_aggregated_phase)
        assert (
            test_num_tokens == num_tokens
        ), f"ERROR! test_num_tokens = {test_num_tokens} not equal to num_tokens = {num_tokens}"

    @pytest.mark.parametrize(
        "phase, stats_aggregated_phase, num_sentences, stats_aggregated_phase_extended",
        [
            (
                "train",
                pd.DataFrame(data=[5, 3, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
                4,
                pd.DataFrame(
                    data={
                        "tags": [5, 3, 0, 0, 0, 0, 0],
                        "tags/sentence": [
                            "{:.2f}".format(x)
                            for x in [5 / 4.0, 3 / 4.0, 0, 0, 0, 0, 0]
                        ],
                        "tags relative w/ 0": [
                            "{:.2f}".format(x)
                            for x in [5 / 8.0, 3 / 8.0, 0, 0, 0, 0, 0]
                        ],
                        "tags relative w/o 0": [
                            "{:.2f}".format(x) for x in [0, 3 / 3.0, 0, 0, 0, 0, 0]
                        ],
                    },
                    index=index,
                    columns=columns,
                ),
            ),
            (
                "val",
                pd.DataFrame(data=[3, 1, 0, 0, 0, 0, 0], index=index, columns=["tags"]),
                2,
                pd.DataFrame(
                    data={
                        "tags": [3, 1, 0, 0, 0, 0, 0],
                        "tags/sentence": [
                            "{:.2f}".format(x)
                            for x in [3 / 2.0, 1 / 2.0, 0, 0, 0, 0, 0]
                        ],
                        "tags relative w/ 0": [
                            "{:.2f}".format(x)
                            for x in [3 / 4.0, 1 / 4.0, 0, 0, 0, 0, 0]
                        ],
                        "tags relative w/o 0": [
                            "{:.2f}".format(x) for x in [0, 1 / 1.0, 0, 0, 0, 0, 0]
                        ],
                    },
                    index=index,
                    columns=columns,
                ),
            ),
        ],
    )
    def test_stats_aggregated_extend(
        self,
        phase: str,
        stats_aggregated_phase: pd.DataFrame,
        num_sentences: int,
        stats_aggregated_phase_extended: pd.DataFrame,
    ):
        test_stats_aggregated_phase_extended = self.analyzer._stats_aggregated_extend(
            stats_aggregated_phase, num_sentences
        )
        pd.testing.assert_frame_equal(
            test_stats_aggregated_phase_extended, stats_aggregated_phase_extended
        ), f"ERROR! test_stats_aggregated_extend pass test for phase = {phase}"

    @pytest.mark.parametrize(
        "stats_aggregated, num_tokens, num_sentences",
        [
            (
                {
                    "train": pd.DataFrame(
                        data={
                            "tags": [5, 3, 0, 0, 0, 0, 0],
                            "tags/sentence": [
                                "{:.2f}".format(x)
                                for x in [5 / 4.0, 3 / 4.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/ 0": [
                                "{:.2f}".format(x)
                                for x in [5 / 8.0, 3 / 8.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/o 0": [
                                "{:.2f}".format(x) for x in [0, 3 / 3.0, 0, 0, 0, 0, 0]
                            ],
                        },
                        index=index,
                        columns=columns,
                    ),
                    "val": pd.DataFrame(
                        data={
                            "tags": [3, 1, 0, 0, 0, 0, 0],
                            "tags/sentence": [
                                "{:.2f}".format(x)
                                for x in [3 / 2.0, 1 / 2.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/ 0": [
                                "{:.2f}".format(x)
                                for x in [3 / 4.0, 1 / 4.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/o 0": [
                                "{:.2f}".format(x) for x in [0, 1 / 1.0, 0, 0, 0, 0, 0]
                            ],
                        },
                        index=index,
                        columns=columns,
                    ),
                    "test": pd.DataFrame(
                        data={
                            "tags": [3, 1, 0, 0, 0, 0, 0],
                            "tags/sentence": [
                                "{:.2f}".format(x)
                                for x in [3 / 2.0, 1 / 2.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/ 0": [
                                "{:.2f}".format(x)
                                for x in [3 / 4.0, 1 / 4.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/o 0": [
                                "{:.2f}".format(x) for x in [0, 1 / 1.0, 0, 0, 0, 0, 0]
                            ],
                        },
                        index=index,
                        columns=columns,
                    ),
                    "total": pd.DataFrame(
                        data={
                            "tags": [11, 5, 0, 0, 0, 0, 0],
                            "tags/sentence": [
                                "{:.2f}".format(x)
                                for x in [11 / 8.0, 5 / 8.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/ 0": [
                                "{:.2f}".format(x)
                                for x in [11 / 16.0, 5 / 16.0, 0, 0, 0, 0, 0]
                            ],
                            "tags relative w/o 0": [
                                "{:.2f}".format(x)
                                for x in [0, 11 / 11.0, 0, 0, 0, 0, 0]
                            ],
                        },
                        index=index,
                        columns=columns,
                    ),
                },
                {"train": 8, "val": 4, "test": 4, "total": 16},
                {"train": 4, "val": 2, "test": 2, "total": 8},
            ),
        ],
    )
    def test_analyze_data(
        self,
        stats_aggregated: Dict[str, pd.DataFrame],
        num_tokens: Dict[str, int],
        num_sentences: Dict[str, int],
    ):
        self.analyzer.analyze_data(
            write_log=False
        )  # created attr: stats_aggregated, num_tokens, num_sentences
        for phase in ["train", "val", "test", "total"]:

            assert self.analyzer.num_tokens[phase] == num_tokens[phase], (
                f"ERROR! test_num_tokens = {self.analyzer.num_tokens[phase]} not equal to "
                f"num_tokens = {num_tokens[phase]} for phase = {phase}"
            )

            assert self.analyzer.num_sentences[phase] == num_sentences[phase], (
                f"ERROR! test_num_sentences = {self.analyzer.num_sentences[phase]} not equal to "
                f"num_sentences = {num_sentences[phase]} for phase = {phase}"
            )

            pd.testing.assert_frame_equal(
                self.analyzer.stats_aggregated[phase], stats_aggregated[phase]
            ), f"ERROR! test_analyze_data / stats_aggregated did not pass test for phase = {phase}"



import subprocess
from typing import List, Dict, Optional, Tuple
import pandas as pd

from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS,
)
from nerblackbox.modules.utils.env_variable import env_variable


class SweNercFormatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "swe_nerc"
        ner_tag_list = ["PRS", "LOC", "GRO", "EVN", "TME", "WRK", "SMP", "MNT"]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]
        """
        bash_cmds = [
            f"mkdir {env_variable('DIR_DATASETS')}/_swe_nerc",
            f"curl -o {env_variable('DIR_DATASETS')}/_swe_nerc/swe_nerc.tar.gz "
            "https://spraakbanken.gu.se/lb/resurser/swe-nerc/Swe-NERC-v1.0.tar.gz",
            f"cd {env_variable('DIR_DATASETS')}/_swe_nerc && tar -xzf swe_nerc.tar.gz",
            f"mkdir {env_variable('DIR_DATASETS')}/swe_nerc/raw_data",
            f"mv {env_variable('DIR_DATASETS')}/_swe_nerc/Swe-NERC-v1.0/manually-tagged-part/*.tsv "
            f"{env_variable('DIR_DATASETS')}/swe_nerc/raw_data",
            f"rm -r {env_variable('DIR_DATASETS')}/_swe_nerc",
            f"echo '\t\t' | tee -a {env_variable('DIR_DATASETS')}/swe_nerc/raw_data/*.tsv",
            #####
            f"cat {env_variable('DIR_DATASETS')}/swe_nerc/raw_data/*-01.tsv "
            f"> {env_variable('DIR_DATASETS')}/swe_nerc/swe_nerc-val.tsv",
            f"cat {env_variable('DIR_DATASETS')}/swe_nerc/raw_data/*-02.tsv "
            f"> {env_variable('DIR_DATASETS')}/swe_nerc/swe_nerc-test.tsv",
            f"cat {env_variable('DIR_DATASETS')}/swe_nerc/raw_data/*-0[!12].tsv "
            f"> {env_variable('DIR_DATASETS')}/swe_nerc/swe_nerc-train.tsv",
            f"cat {env_variable('DIR_DATASETS')}/swe_nerc/raw_data/*-[!0]?.tsv "
            f">> {env_variable('DIR_DATASETS')}/swe_nerc/swe_nerc-train.tsv",
        ]

        for bash_cmd in bash_cmds:
            if verbose:
                print(bash_cmd)

            try:
                subprocess.run(bash_cmd, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                print(e)

    def create_ner_tag_mapping(self) -> Dict[str, str]:
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            sentences_rows_iob2: only if write_csv = False
        """
        for phase in ["train", "val", "test"]:
            sentences_rows_iob1 = self._read_original_file(phase)
            if shuffle:
                sentences_rows_iob1 = self._shuffle_dataset(phase, sentences_rows_iob1)

            sentences_rows_iob2 = self._convert_iob1_to_iob2(sentences_rows_iob1)
            if write_csv:  # pragma: no cover
                self._write_formatted_csv(phase, sentences_rows_iob2)
            else:
                return sentences_rows_iob2
        return None

    def set_original_file_paths(self) -> None:
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        self.file_name = {
            phase: f"swe_nerc-{phase}.tsv" for phase in ["train", "val", "test"]
        }

    def _parse_row(self, _row: str) -> List[str]:
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        _row_list_raw = _row.replace("\t", " ").strip().split(" ")
        if (len(_row_list_raw) == 4 and _row_list_raw[-1] == "B") or (
            len(_row_list_raw) == 3
        ):
            return _row_list_raw
        elif len(_row_list_raw) > 2:
            return [" ".join(_row_list_raw[:-2]), _row_list_raw[-2], _row_list_raw[-1]]
        elif _row in ["\n", "\t\t\n"] or len(_row) == 0:
            return []
        elif _row.startswith("\t"):
            print(f"ATTENTION! row = {repr(_row)} observed -> will be skipped")
            return ["SKIP-THIS-TOKEN", "0", "PAD"]
        else:  # pragma: no cover
            raise Exception(f"ERROR! could not parse row = {repr(_row)}")

    def _format_original_file(self, _row_list: List[str]) -> Optional[List[str]]:
        """
        III: format data

        Args:
            _row_list: e.g. ["test", "PER", "X", "B"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        if not len(_row_list) in [3, 4]:  # pragma: no cover
            raise Exception(
                f"ERROR! row_list = {_row_list} should consist of 3 or 4 parts!"
            )

        if _row_list[0] != "SKIP-THIS-TOKEN":
            # transformation on _row_list[0] replaces unwanted nbsp characters
            _row_list_formatted = [
                "".join(_row_list[0].split()),
                self.transform_tags(_row_list),
            ]
            return _row_list_formatted
        else:
            return None

    def resplit_data(
        self, val_fraction: float = 0.0, write_csv: bool = True
    ) -> Optional[Tuple[pd.DataFrame, ...]]:
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            df_train: only if write_csv = False
            df_val:   only if write_csv = False
            df_test:  only if write_csv = False
        """
        # train -> train
        df_train = self._read_formatted_csvs(["train"])

        # val  -> val
        df_val = self._read_formatted_csvs(["val"])

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])

        if write_csv:  # pragma: no cover
            self._write_final_csv("train", df_train)
            self._write_final_csv("val", df_val)
            self._write_final_csv("test", df_test)
            return None
        else:
            return df_train, df_val, df_test

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    @staticmethod
    def transform_tags(_row_list: List[str]) -> str:
        """
        Args:
            _row_list: ["test", "O", "XYZ"], ["test-2", "PER", "XYZ"] or ["test-3", "PER", "XYZ", "B"]

        Return:
            tag: e.g. "O", "I-PER" or "B-PER"
        """
        assert len(_row_list) in [
            3,
            4,
        ], f"ERROR! encountered row_list = {_row_list} that cannot be parsed."
        plain_tag = _row_list[1]
        if plain_tag == "O":
            return plain_tag
        else:
            if len(_row_list) == 3:
                return f"I-{plain_tag}"
            elif _row_list[3] == "B":
                return f"B-{plain_tag}"
            else:
                raise Exception(
                    f"ERROR! encountered row_list = {_row_list} that cannot be parsed."
                )



from typing import List, Dict, Optional, Tuple, Any
import pandas as pd
from os.path import join
from datasets import (
    load_dataset,
    load_dataset_builder,
    get_dataset_split_names,
    DatasetDict,
)
from shutil import copyfile

from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS,
    SENTENCES_ROWS_PRETOKENIZED,
    SENTENCES_ROWS_UNPRETOKENIZED,
)


class HuggingfaceDatasetsFormatter(BaseFormatter):

    PHASES = ["train", "val", "test"]
    PHASES_DATASETS = {"train": "train", "val": "validation", "test": "test"}

    @classmethod
    def check_existence(
        cls, ner_dataset: str, ner_dataset_subset: str = ""
    ) -> Tuple[bool, str]:
        """
        checks if ner_dataset exists in huggingface datasets

        Args:
            ner_dataset: e.g. "conll2003"
            ner_dataset_subset: e.g. "simple_cased"

        Returns:
            existence: True if ner_dataset exists in huggingface datasets, False otherwise
        """
        try:
            _ = load_dataset_builder(
                ner_dataset,
                name=ner_dataset_subset if len(ner_dataset_subset) else None,
            )
            return True, ""
        except ValueError as e:
            return (
                False,
                f"Error! config name is missing for ner_dataset = {ner_dataset} "
                f"(ner_dataset_subset = {ner_dataset_subset})! Error message: {e}",
            )
        except FileNotFoundError:
            return False, f"Error! ner_dataset = {ner_dataset} unknown."

    @classmethod
    def check_compatibility(
        cls, ner_dataset: str, ner_dataset_subset: str = ""
    ) -> Tuple[bool, str]:
        """
        checks if ner_dataset contains train/val/test splits

        Args:
            ner_dataset: e.g. "conll2003"
            ner_dataset_subset: e.g. "simple_cased"

        Returns:
            compatibility: True if ner_dataset contains train/val/test splits, False otherwise
        """
        _dataset_split_names = get_dataset_split_names(
            ner_dataset,
            config_name=ner_dataset_subset if len(ner_dataset_subset) else None,
        )
        compatibility = sorted(_dataset_split_names) == ["test", "train", "validation"]
        error_msg = (
            ""
            if compatibility
            else f"ner_dataset = {ner_dataset} does not contain train/val/test splits."
        )
        return compatibility, error_msg

    @classmethod
    def check_implementation(
        cls, ner_dataset: str, ner_dataset_subset: str = ""
    ) -> Tuple[bool, str]:
        """
        problem: there is no common structure in dataset_builder.info.features for all datasets
        parsing for a few typical structures is implemented
        this method checks if one of them is applicable to ner_dataset

        Args:
            ner_dataset: e.g. "conll2003"
            ner_dataset_subset: e.g. "simple_cased"

        Returns:
            implementation: True if ner_dataset is implemented, False otherwise
        """
        implementation, _, _, _ = cls.get_infos(ner_dataset, ner_dataset_subset)
        error_msg = (
            "" if implementation else f"ner_dataset = {ner_dataset} can not be parsed."
        )
        return implementation, error_msg

    @classmethod
    def get_infos(
        cls,
        ner_dataset: str,
        ner_dataset_subset: str = "",
    ) -> Tuple[bool, Optional[List[str]], Optional[bool], Optional[Dict[str, Any]]]:
        """
        get all relevant infos about dataset

        Args:
            ner_dataset: e.g. "conll2003"
            ner_dataset_subset: e.g. "simple_cased"

        Returns:
            implementation: True if ner_dataset is implemented, False otherwise
            tags: e.g. ["O", "B-LOC", "B-MISC", "B-ORG", "B-PER", "I-LOC", "I-MISC", "I-ORG", "I-PER"]
            pretokenized: e.g. True
            lookup_table: e.g. {'text': 'tokens', 'tags': 'ner_tags', 'mapping': None}
                          e.g. {'text': 'sentence', 'tags': 'entities', 'mapping': {..}}
        """
        dataset_builder = load_dataset_builder(
            ner_dataset, name=ner_dataset_subset if len(ner_dataset_subset) else None
        )
        if dataset_builder.info.features is None:
            return False, None, None, None
        else:
            feat = dict(dataset_builder.info.features)

        implementation: bool = False
        tags: Optional[List[str]] = None
        pretokenized: Optional[bool] = None
        lookup_table: Optional[Dict[str, Any]] = None
        try:
            if "ner_tags" in feat:  # e.g. conll2003
                keys = ["tokens", "ner_tags"]
                if all([key in feat for key in keys]):
                    implementation = True
                    tags = feat["ner_tags"].feature.names
                    pretokenized = True
                    lookup_table = {
                        "text": "tokens",
                        "tags": "ner_tags",
                        "mapping": None,
                    }
            elif "entities" in feat:  # e.g. ehealth_kd
                keys = ["sentence", "entities"]
                if all([key in feat for key in keys]):
                    entities_keys = [
                        "ent_text",
                        "ent_label",
                        "start_character",
                        "end_character",
                    ]
                    if all(
                        [
                            entities_key in feat["entities"][0]
                            for entities_key in entities_keys
                        ]
                    ):
                        implementation = True
                        tags = feat["entities"][0]["ent_label"].names
                        pretokenized = False
                        lookup_table = {
                            "text": "sentence",
                            "tags": "entities",
                            "mapping": {
                                "ent_text": "token",
                                "ent_label": "tag",
                                "start_character": "char_start",
                                "end_character": "char_end",
                            },
                        }
            else:
                return False, None, None, None
        except Exception:
            return False, None, None, None

        if (
            implementation is False
            or tags is None
            or pretokenized is None
            or lookup_table is None
        ):
            return False, None, None, None
        else:
            return implementation, tags, pretokenized, lookup_table

    def __init__(self, ner_dataset: str, ner_dataset_subset: str = ""):
        _, self.tags, self.pretokenized, self.lookup_table = self.get_infos(
            ner_dataset, ner_dataset_subset
        )
        self.sentences_rows_pretokenized: Dict[str, SENTENCES_ROWS_PRETOKENIZED] = {
            phase: list() for phase in self.PHASES
        }
        self.sentences_rows_unpretokenized: Dict[str, SENTENCES_ROWS_UNPRETOKENIZED] = {
            phase: list() for phase in self.PHASES
        }
        super().__init__(ner_dataset, ner_tag_list=self.get_ner_tag_list())
        self.ner_dataset_subset = ner_dataset_subset

    def get_ner_tag_list(self) -> List[str]:
        """
        reduces tags to plain tags

        Used attr:
            tags: e.g. ["O", "B-LOC", "B-MISC", "B-ORG", "B-PER", "I-LOC", "I-MISC", "I-ORG", "I-PER"]

        Returns:
            ner_tag_list: ordered, e.g.["LOC", "MISC", "ORG", "PER"]
        """
        assert self.tags is not None, f"ERROR! self.tags unexpectedly found to be None."
        return sorted(
            list(set([tag.split("-")[-1] for tag in self.tags if tag != "O"]))
        )

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]

        Created attr:
            sentences_rows: e.g. (-pretokenized-)
                            [
                                [['Inger', 'PER'], ['säger', '0'], .., []],
                                [['Det', '0'], .., []]
                            ]
                            e.g. (-not pretokenized-)
                            [
                                {
                                    'text': 'Inger säger ..',
                                    'tags': [{'token': 'Inger', 'tag': 'PER', 'char_start': 0, 'char_end': 5}, ..],
                                },
                                {
                                    'text': 'Det ..',
                                    'tags': [{..}, ..]
                                }
                            ]
        """
        # typing
        assert self.tags is not None, f"ERROR! self.tags unexpectedly found to be None."
        assert (
            self.lookup_table is not None
        ), f"ERROR! self.lookup_table unexpectedly found to be None."

        # start
        text = self.lookup_table["text"]
        tags = self.lookup_table["tags"]
        dataset = load_dataset(
            self.ner_dataset,
            self.ner_dataset_subset if len(self.ner_dataset_subset) else None,
        )
        assert isinstance(
            dataset, DatasetDict
        ), f"ERROR! type(dataset) = {type(dataset)} should be DatasetDict"
        for phase in self.PHASES:
            phase_dataset = self.PHASES_DATASETS[phase]
            for field in [text, tags]:
                assert (
                    field in dataset[phase_dataset].info.features.keys()
                ), f"ERROR! field = {field} not present in dataset."
            for text_sentence, tags_sentence in zip(
                dataset[phase_dataset][:][text], dataset[phase_dataset][:][tags]
            ):
                if self.pretokenized:
                    self.sentences_rows_pretokenized[phase].append(
                        [
                            [text_single, self.tags[int(tag_single)]]
                            for text_single, tag_single in zip(
                                text_sentence, tags_sentence
                            )
                        ]
                    )
                else:
                    _dict = {
                        "text": text_sentence,
                        "tags": [
                            {
                                self.lookup_table["mapping"][key]: value
                                for key, value in tags_sentence[n].items()
                                if key in self.lookup_table["mapping"]
                            }
                            for n in range(len(tags_sentence))
                        ],
                    }
                    for tag_dict in _dict["tags"]:
                        tag_dict.update(
                            (k, self.tags[int(v)])
                            for k, v in tag_dict.items()
                            if k == "tag"
                        )

                    # there are multiword entities that are not connected
                    # (i.e. there are other words between the entities' tokens)
                    # e.g. ehealth_kd: {"token": "uno días", "tag": "Concept", "char_start": 64170, "char_end": 64183}
                    # -> filter them out
                    _dict["tags"] = [
                        elem
                        for elem in _dict["tags"]
                        if len(elem["token"]) == elem["char_end"] - elem["char_start"]
                    ]

                    self.sentences_rows_unpretokenized[phase].append(_dict)

    def create_ner_tag_mapping(self) -> Dict[str, str]:
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            sentences_rows_iob2: only if write_csv = False
        """
        if self.pretokenized:
            for phase in ["train", "val", "test"]:
                sentences_rows_pretokenized_phase = self.sentences_rows_pretokenized[
                    phase
                ].copy()
                if shuffle:
                    sentences_rows_pretokenized_phase = self._shuffle_dataset(
                        phase, sentences_rows_pretokenized_phase
                    )

                sentences_rows_iob2 = self._convert_iob1_to_iob2(
                    sentences_rows_pretokenized_phase
                )
                if write_csv:  # pragma: no cover
                    self._write_formatted_csv(phase, sentences_rows_iob2)
                else:
                    return sentences_rows_iob2  # return train!
        else:
            for phase in ["train", "val", "test"]:
                sentences_rows_unpretokenized_phase = (
                    self.sentences_rows_unpretokenized[phase].copy()
                )
                if shuffle:
                    sentences_rows_unpretokenized_phase = self._shuffle_dataset(
                        phase, sentences_rows_unpretokenized_phase
                    )

                if write_csv:  # pragma: no cover
                    self._write_formatted_jsonl(
                        phase, sentences_rows_unpretokenized_phase
                    )
                else:
                    return sentences_rows_unpretokenized_phase  # returns train!
        return None

    def set_original_file_paths(self) -> None:  # pragma: no cover
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        pass  # not necessary as _read_original_file() is not used

    def _parse_row(self, _row: str) -> List[str]:  # pragma: no cover
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        # not necessary as _read_original_file() is not used
        return []

    def _format_original_file(
        self, _row_list: List[str]
    ) -> Optional[List[str]]:  # pragma: no cover
        """
        III: format data

        Args:
            _row_list: e.g. ["test", "PER", "X", "B"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        pass  # not necessary as _read_original_file() is not used

    def resplit_data(
        self, val_fraction: float = 0.0, write_csv: bool = True
    ) -> Optional[Tuple[pd.DataFrame, ...]]:
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            df_train: only if write_csv = False
            df_val:   only if write_csv = False
            df_test:  only if write_csv = False
        """
        if self.pretokenized:
            # train -> train
            df_train = self._read_formatted_csvs(["train"])

            # val  -> val
            df_val = self._read_formatted_csvs(["val"])

            # test  -> test
            df_test = self._read_formatted_csvs(["test"])

            if write_csv:  # pragma: no cover
                self._write_final_csv("train", df_train)
                self._write_final_csv("val", df_val)
                self._write_final_csv("test", df_test)
                return None
            else:
                return df_train, df_val, df_test
        else:  # pragma: no cover
            for phase in self.PHASES:
                src = join(self.dataset_path, f"{phase}_formatted.jsonl")
                dst = join(self.dataset_path, f"{phase}.jsonl")
                copyfile(src, dst)
            return None



from pytorch_lightning.callbacks.early_stopping import EarlyStopping


class CustomEarlyStopping(EarlyStopping):
    def __init__(
        self,
        lr_schedule: str,
        lr_cooldown_epochs: int,
        lr_cooldown_restarts: bool,
        *args,
        **kwargs,
    ):
        self.lr_schedule = lr_schedule
        self.lr_cooldown_epochs = lr_cooldown_epochs
        self.lr_cooldown_restarts = lr_cooldown_restarts
        super().__init__(*args, **kwargs)

        self.use_cooldown = (
            self.lr_schedule == "constant" and self.lr_cooldown_epochs > 0
        )
        # e.g. patience = 4, lr_cooldown_epochs = 3 -> wait_count_trigger_cool_down = 1
        self.wait_count_trigger_cool_down = self.patience - self.lr_cooldown_epochs
        self.flag_cooldown = False  # Custom

    def on_validation_end(self, trainer, pl_module):
        # Original
        if self._check_on_train_epoch_end or self._should_skip_check(trainer):
            return

        self._run_early_stopping_check(trainer)

        # Custom
        if self.use_cooldown:
            if self.wait_count == self.wait_count_trigger_cool_down:
                # NO IMPROVEMENT
                print(
                    f"> Custom Early Stopping: change LR to linear cool-down with {self.lr_cooldown_epochs} epochs."
                )
                pl_module.scheduler = pl_module._create_scheduler(
                    _lr_warmup_epochs=0,
                    _lr_schedule="linear",
                    _lr_max_epochs=self.lr_cooldown_epochs,
                )
                self.flag_cooldown = True
                if self.lr_cooldown_restarts is False:
                    # => suppress reoccurrence of improvement
                    self.min_delta = -1000000 if self.mode == "min" else 1000000
            elif (
                self.lr_cooldown_restarts
                and self.flag_cooldown is True
                and self.wait_count == 0
            ):
                # REOCCURRENCE OF IMPROVEMENT
                print(f"> Custom Early Stopping: change LR to normal constant mode.")
                pl_module.scheduler = pl_module._create_scheduler(
                    _lr_warmup_epochs=0,
                    _lr_schedule="constant",
                )
                self.flag_cooldown = False



import pytest
from typing import List
from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics


class TestLoggedMetrics:

    logged_metrics = LoggedMetrics()

    @pytest.mark.parametrize(
        "required_tag_groups, "
        "required_phases, "
        "required_levels, "
        "required_averaging_groups, "
        "exclude, "
        "metrics",
        [
            (
                ["all"],
                ["train"],
                None,
                None,
                [
                    "numberofclasses",
                    "asr_precision",
                    "asr_recall",
                    "asr_f1",
                    "asr_abidance",
                ],
                ["loss", "acc"],
            ),
            (
                ["all"],
                ["train"],
                None,
                None,
                ["numberofclasses", "acc"],
                ["loss"],
            ),
            (
                ["ind"],
                ["test"],
                None,
                None,
                ["numberofclasses"],
                ["precision", "recall", "f1"],
            ),
            (
                ["ind"],
                ["test"],
                None,
                ["micro"],
                ["numberofclasses"],
                ["precision", "recall", "f1"],
            ),
            (
                ["ind"],
                ["val"],
                None,
                ["micro"],
                ["numberofclasses"],
                [],
            ),
            (
                ["ind"],
                ["test"],
                None,
                ["macro"],
                ["numberofclasses"],
                [],
            ),
            (
                None,
                None,
                ["entity"],
                None,
                [
                    "numberofclasses",
                    "asr_precision",
                    "asr_recall",
                    "asr_f1",
                    "asr_abidance",
                ],
                ["precision", "recall", "f1"],
            ),
            (
                None,
                None,
                None,
                None,
                [
                    "numberofclasses",
                    "asr_precision",
                    "asr_recall",
                    "asr_f1",
                    "asr_abidance",
                ],
                ["loss", "acc", "precision", "recall", "f1"],
            ),
            (
                None,
                None,
                None,
                None,
                None,
                [
                    "loss",
                    "acc",
                    "precision",
                    "recall",
                    "f1",
                    "numberofclasses",
                    "asr_precision",
                    "asr_recall",
                    "asr_f1",
                    "asr_abidance",
                ],
            ),
            (
                None,
                None,
                ["entity"],
                ["macro"],
                None,
                ["precision", "recall", "f1", "numberofclasses"],
            ),
            (
                None,
                None,
                ["entity"],
                ["micro"],
                None,
                ["precision", "recall", "f1", "asr_precision", "asr_recall", "asr_f1"],
            ),
            (
                None,
                None,
                ["entity"],
                ["simple"],
                None,
                ["asr_abidance"],
            ),
        ],
    )
    def test_get_metrics(
        self,
        required_tag_groups: List[str],
        required_phases: List[str],
        required_levels: List[str],
        required_averaging_groups: List[str],
        exclude: List[str],
        metrics: List[str],
    ):

        test_metrics = self.logged_metrics.get_metrics(
            required_tag_groups=required_tag_groups,
            required_phases=required_phases,
            required_levels=required_levels,
            required_averaging_groups=required_averaging_groups,
            exclude=exclude,
        )
        assert set(test_metrics) == set(
            metrics
        ), f"test_metrics = {test_metrics} != {metrics}"

    @pytest.mark.parametrize(
        "metrics_flat_list",
        [
            (
                [
                    "entity_fil_numberofclasses_macro",
                    "entity_fil_asr_precision_micro",
                    "entity_fil_asr_recall_micro",
                    "entity_fil_asr_f1_micro",
                    "entity_fil_asr_abidance",
                    "entity_fil_precision_micro",
                    "entity_fil_precision_macro",
                    "entity_fil_recall_micro",
                    "entity_fil_recall_macro",
                    "entity_fil_f1_micro",
                    "entity_fil_f1_macro",
                    #
                    "entity_ind_precision_micro",
                    "entity_ind_recall_micro",
                    "entity_ind_f1_micro",
                    #
                    "token_all_loss",
                    "token_all_acc",
                    #
                    "token_all_precision_micro",
                    "token_all_precision_macro",
                    "token_all_recall_micro",
                    "token_all_recall_macro",
                    "token_all_f1_micro",
                    "token_all_f1_macro",
                    #
                    "token_fil_numberofclasses_macro",
                    "token_fil_precision_micro",
                    "token_fil_precision_macro",
                    "token_fil_recall_micro",
                    "token_fil_recall_macro",
                    "token_fil_f1_micro",
                    "token_fil_f1_macro",
                    #
                    "token_ind_precision_micro",
                    "token_ind_recall_micro",
                    "token_ind_f1_micro",
                    #
                    "token_O_precision_micro",
                    "token_O_recall_micro",
                    "token_O_f1_micro",
                ]
            )
        ],
    )
    def test_as_flat_list(self, metrics_flat_list: List[str]):
        test_metrics_flat_list = self.logged_metrics.as_flat_list()
        assert set(test_metrics_flat_list) == set(
            metrics_flat_list
        ), f"test_metrics_flat_list = {test_metrics_flat_list} != {metrics_flat_list}"



import os
import glob
from typing import List, Union, Dict, Tuple, Optional, Any
from os.path import isdir, abspath, join, isfile
import subprocess
import signal
from pkg_resources import Requirement
from pkg_resources import resource_filename, resource_isdir
import shutil
import pandas as pd
from mlflow.tracking import MlflowClient
from mlflow.entities import Run

from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.experiment_results import ExperimentResults


VERBOSE = True


class Store:
    r"""
    client for the store that contains all data (datasets, experiment configuration files, models, results)

    Attributes:
         path: path to store's main directory
    """
    path: Optional[str] = os.environ.get("DATA_DIR")

    # will be set by _update_client() & _update_experiments()
    client: Optional[MlflowClient] = None
    experiment_id2name: Optional[Dict[str, str]] = None
    experiment_name2id: Optional[Dict[str, str]] = None

    process: Dict[str, Any] = {
        "mlflow": None,
        "tensorboard": None,
    }

    @classmethod
    def get_path(cls) -> Optional[str]:
        r"""
        Returns:
            cls.path: path to store's main directory
        """
        return cls.path

    @classmethod
    def set_path(cls, path: str) -> Optional[str]:
        r"""
        Args:
            path: path to store's main directory
        Returns:
            cls.path: path to store's main directory
        """
        data_dir = abspath(path)
        os.environ["DATA_DIR"] = data_dir  # environment variable
        cls.path = data_dir  # cls.path
        cls._update_client()  # cls.client
        return cls.get_path()

    @classmethod
    def create(cls) -> None:
        r"""
        create store at cls.path
        """
        if resource_isdir(Requirement.parse("nerblackbox"), "nerblackbox/modules/data"):
            data_source = resource_filename(
                Requirement.parse("nerblackbox"), "nerblackbox/modules/data"
            )

            data_dir = env_variable("DATA_DIR")
            if VERBOSE:
                print("data_source =", data_source)
                print("data_target =", data_dir)

            if isdir(data_dir):
                print(f"> {cls.path} exists.")
            else:
                shutil.copytree(data_source, data_dir)
                print(f"> {cls.path} created.")

            cls._update_client()
        else:
            print("Store.create() not executed successfully")
            exit(0)

    @classmethod
    def show_experiments(
        cls, as_df: bool = True
    ) -> Union[pd.DataFrame, Dict[str, str]]:
        r"""
        Args:
            as_df: if True, return pandas DataFrame. if False, return dict

        Returns:
            experiments: overview of experiments that have been run
        """
        cls._update_experiments()

        assert isinstance(
            cls.experiment_id2name, dict
        ), f"ERROR! type(cls.experiment_id2name) = {type(cls.experiment_id2name)} should be dict."

        experiments_overview = sorted(
            [
                {"experiment_id": k, "experiment_name": v}
                for k, v in cls.experiment_id2name.items()
            ],
            key=lambda elem: elem["experiment_id"],
        )
        return pd.DataFrame(experiments_overview) if as_df else experiments_overview

    @classmethod
    def get_experiment_results(cls) -> List[ExperimentResults]:
        r"""
        get results for all experiments

        Returns:
            experiment_results_list: TODO: instead of list return dict that maps experiment_name to ExperimentResults?
        """
        cls._update_experiments()

        assert (
            cls.experiment_name2id is not None
        ), f"ERROR! cls.experiment_name2id is None."

        # get ExperimentResults
        experiment_results_list: List[ExperimentResults] = list()
        for _name, _id in sorted(
            list(cls.experiment_name2id.items()), key=lambda x: x[1]
        ):  # sort by id
            experiment_exists, experiment_results = cls.get_experiment_results_single(
                _name, update_experiments=False
            )
            assert experiment_exists, f"ERROR! experiment = {_name} does not exist."
            experiment_results_list.append(experiment_results)

        # return
        # CLI skipped
        return experiment_results_list

    @classmethod
    def get_experiment_results_single(
        cls, experiment_name: str, update_experiments: bool = True
    ) -> Tuple[bool, ExperimentResults]:
        r"""
        get results for single experiment

        Args:
            experiment_name: e.g. 'exp0'
            update_experiments: whether to update cls.experiment_id2name & cls.experiment_name2id

        Returns:
            experiment_results: for experiment with experiment_name
        """
        if cls.client is None:
            cls._update_client()

        if isdir(env_variable("DIR_MLFLOW")):
            if update_experiments:
                cls._update_experiments()

            assert (
                cls.experiment_name2id is not None
            ), f"ERROR! cls.experiment_name2id is None."

            if experiment_name in cls.experiment_name2id.keys():
                experiment_id: str = cls.experiment_name2id[experiment_name]
                assert isinstance(
                    cls.client, MlflowClient
                ), f"ERROR! type(cls.client) = {type(cls.client)} should be MlflowClient"
                runs: List[Run] = cls.client.search_runs([experiment_id])

                return True, ExperimentResults.from_mlflow_runs(
                    runs,
                    experiment_id,
                    experiment_name,
                )
            else:
                print(f"no experiment with experiment_name = {experiment_name} found")
                print(f"experiments that were found:")
                print(list(cls.experiment_name2id.keys()))
                return False, ExperimentResults()
        else:
            return False, ExperimentResults()

    @classmethod
    def mlflow(cls, action: str) -> None:
        """
        start or stop the mlflow server at http://127.0.0.1:5000 or check its status

        Args:
            action: "start", "status", "stop"
        """
        cls._subprocess(action, "mlflow")

    @classmethod
    def tensorboard(cls, action: str) -> None:
        """
        start or stop the tensorboard server at http://127.0.0.1:6006 or check its status

        Args:
            action: "start", "status", "stop"
        """
        cls._subprocess(action, "tensorboard")

    @classmethod
    def clear_data(cls, results: bool = False) -> None:
        """
        :used attr: clear_all [bool] if True, clear not only checkpoints but also mlflow, tensorboard and logs
        """
        assert isinstance(
            cls.path, str
        ), f"ERROR! type(cls.path) = {type(cls.path)} should be str."
        results_dir = join(cls.path, "results")
        assert isdir(results_dir), f"directory {results_dir} does not exist."

        # checkpoints
        objects_to_remove = glob.glob(join(results_dir, "checkpoints", "*"))  # list

        # results (mlflow, tensorboard, ..)
        if results:
            results_files = (
                glob.glob(join(results_dir, "mlruns", "*"))
                + glob.glob(join(results_dir, "mlruns", ".*"))
                + glob.glob(join(results_dir, "tensorboard", "*"))
                + glob.glob(join(results_dir, "logs.log"))
                + glob.glob(join(results_dir, "*.npy"))
            )
            objects_to_remove.extend(results_files)

        if len(objects_to_remove) == 0:
            print(f"There is no data to remove in {results_dir}")
        else:
            for elem in objects_to_remove:
                print(elem)
            while 1:
                answer = input("Do you want to remove the above files? (y/n) ")
                if answer == "y":
                    for elem in objects_to_remove:
                        if isfile(elem):
                            os.remove(elem)
                        elif isdir(elem):
                            shutil.rmtree(elem, ignore_errors=False)
                        else:
                            raise ValueError(
                                f"object {elem} is neither a file nor a dir and cannot be removed"
                            )
                    print(f"Files removed")
                    break
                elif answer == "n":
                    print(f"No files removed")
                    break
                else:
                    print("Please enter either y or n")

    ####################################################################################################################
    # HELPER METHODS
    ####################################################################################################################
    @classmethod
    def _update_client(cls) -> None:
        """
        Changed Attr:
            client: [MlflowClient]
        """
        os.environ["MLFLOW_TRACKING_URI"] = env_variable("DIR_MLFLOW")

        data_dir_exists_before = isdir(env_variable("DATA_DIR"))
        mlflow_subdirectory_exists_before = isdir(env_variable("DIR_MLFLOW"))
        cls.client = (
            MlflowClient()
        )  # creates initial subdirectory DATA_DIR/results/mlruns/0 (if cli is used)
        mlflow_subdirectory_exists_after = isdir(env_variable("DIR_MLFLOW"))

        if (
            mlflow_subdirectory_exists_before is False
            and mlflow_subdirectory_exists_after is True
        ):
            if data_dir_exists_before is False:
                # if whole DATA_DIR is new => delete whole DATA_DIR
                shutil.rmtree(env_variable("DATA_DIR"))
            else:
                # if only DIR_MLFLOW is new => delete only DIR_MLFLOW
                shutil.rmtree(env_variable("DIR_MLFLOW"))

        if data_dir_exists_before:
            # create DIR_MLFLOW and subdirectory .trash
            os.makedirs(join(env_variable("DIR_MLFLOW"), ".trash"), exist_ok=True)

    @classmethod
    def _update_experiments(cls) -> None:
        """
        Changed Attr:
            experiment_id2name: [dict] w/ keys = experiment_id   [str] & values = experiment_name [str]
            experiment_name2id: [dict] w/ keys = experiment_name [str] & values = experiment_id   [str]
        """
        assert cls.client is not None, f"ERROR! cls.client is None."
        cls.experiment_id2name = {
            elem["_experiment_id"]: elem["_name"]
            for elem in [
                vars(experiment) for experiment in cls.client.search_experiments()
            ]
            if elem["_name"] != "Default"
        }

        assert (
            cls.experiment_id2name is not None
        ), f"ERROR! cls.experiment_id2name is None."
        cls.experiment_name2id = {v: k for k, v in cls.experiment_id2name.items()}

    @classmethod
    def _subprocess(cls, _action: str, server_type: str) -> None:
        """
        start or stop a mlflow/tensorboard subprocess, or check its status
        see https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true

        Args:
            _action: "start", "status", "stop"
            server_type: "mlflow", "tensorboard"
        """
        link = cls._get_link(server_type)
        if _action == "start":
            if cls.process[server_type] is None:
                start_command = cls._get_start_command(server_type)
                cls.process[server_type] = subprocess.Popen(
                    start_command,
                    stdout=subprocess.PIPE,
                    shell=True,
                    preexec_fn=os.setsid,
                )
                print(
                    f"{server_type} process with pid = {cls.process[server_type].pid} started"
                )
                print(link)
            else:
                print(
                    f"{server_type} process with pid = {cls.process[server_type].pid} already running"
                )
                print(link)
        elif _action == "status":
            if cls.process[server_type] is None:
                print(f"no {server_type} process found")
            else:
                print(
                    f"{server_type} process with pid = {cls.process[server_type].pid} running"
                )
                print(link)
        elif _action == "stop":
            if cls.process[server_type] is None:
                print(f"no {server_type} process found")
            else:
                os.killpg(os.getpgid(cls.process[server_type].pid), signal.SIGTERM)
                print(
                    f"{server_type} process with pid = {cls.process[server_type].pid} killed"
                )
                print(link)
                cls.process[server_type] = None

    @classmethod
    def _get_start_command(cls, server_type: str) -> str:
        """
        Args:
            server_type: "mlflow", "tensorboard"

        Returns:
            start_command: e.g. "cd [..]; mlflow ui"
        """
        cmd_cd = f'cd {join(env_variable("DATA_DIR"), "results")}'
        if server_type == "mlflow":
            return f"{cmd_cd}; mlflow ui"
        elif server_type == "tensorboard":
            return f"{cmd_cd}; tensorboard --logdir tensorboard --reload_multifile=true"
        else:
            raise Exception(f"ERROR! server_type = {server_type} unknown.")

    @classmethod
    def _get_link(cls, server_type: str) -> str:
        """
        Args:
            server_type: "mlflow", "tensorboard"

        Returns:
            link: e.g. "see http://127.0.0.1:5000"
        """
        if server_type == "mlflow":
            return f"see http://127.0.0.1:5000"
        elif server_type == "tensorboard":
            return f"see http://127.0.0.1:6006"
        else:
            raise Exception(f"ERROR! server_type = {server_type} unknown.")



import pytest
from typing import Optional, Dict, Any
from nerblackbox.api.utils import Utils


class TestApiUtils:

    # 1 ################################################################################################################
    @pytest.mark.parametrize(
        "kwargs_optional, kwargs",
        [
            (
                None,
                {},
            ),
            (
                {"a": 1, "b": None},
                {"a": 1},
            ),
        ],
    )
    def test_process_kwargs_optional(
        self, kwargs_optional: Optional[Dict[str, Any]], kwargs: Dict[str, Any]
    ):
        test_kwargs = Utils.process_kwargs_optional(kwargs_optional)
        assert test_kwargs == kwargs, f"ERROR! test_kwargs = {test_kwargs} != {kwargs}"

    # 2 ################################################################################################################
    @pytest.mark.parametrize(
        "_kwargs, _hparams",
        [
            (
                {},
                {},
            ),
            (
                {"fp16": True},
                {},
            ),
            (
                {"a": 1, "fp16": None},
                {"a": 1},
            ),
        ],
    )
    def test_extract_hparams(self, _kwargs: Dict[str, Any], _hparams: Dict[str, Any]):
        test_hparams = Utils.extract_hparams(_kwargs)
        assert (
            test_hparams == _hparams
        ), f"ERROR! test_hparams = {test_hparams} != {_hparams}"



import os
from os.path import join, isfile
import subprocess
import json
import pandas as pd
from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Optional, Union, Any
import random

from nerblackbox.modules.utils.util_functions import get_dataset_path
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.datasets.formatter.util_functions import get_ner_tag_mapping
from nerblackbox.modules.datasets.analyzer import Analyzer

SEED_SHUFFLE = {
    "train": 4,
    "val": 5,
    "test": 6,
}

SENTENCES_ROWS_PRETOKENIZED = List[List[List[str]]]
SENTENCES_ROWS_UNPRETOKENIZED = List[Dict[str, Any]]
SENTENCES_ROWS = Union[SENTENCES_ROWS_PRETOKENIZED, SENTENCES_ROWS_UNPRETOKENIZED]


class BaseFormatter(ABC):
    def __init__(
        self, ner_dataset: str, ner_tag_list: List[str], ner_dataset_subset: str = ""
    ):
        """
        Args:
            ner_dataset:  'swedish_ner_corpus' or 'suc'
            ner_tag_list: e.g. ['PER', 'LOC', ..]
            ner_dataset_subset: e.g. 'original_cased'
        """
        self.ner_dataset: str = ner_dataset
        self.ner_tag_list: List[str] = ner_tag_list
        self.dataset_path: str = get_dataset_path(ner_dataset, ner_dataset_subset)
        self.file_name: Dict[str, str] = {}
        self.analyzer = Analyzer(self.ner_dataset, self.ner_tag_list, self.dataset_path)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    @abstractmethod
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]
        """
        pass

    @abstractmethod
    def create_ner_tag_mapping(self) -> Dict[str, str]:  # pragma: no cover
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        pass

    @abstractmethod
    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:  # pragma: no cover
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)
        """
        pass

    def set_original_file_paths(self) -> None:  # pragma: no cover
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        pass

    @abstractmethod
    def _parse_row(self, _row: str) -> List[str]:  # pragma: no cover
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        pass

    def _format_original_file(
        self, _row_list: List[str]
    ) -> Optional[List[str]]:  # pragma: no cover
        """
        III: format data

        Args:
            _row_list: e.g. ["test", "PER", "X", "B"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        pass

    @abstractmethod
    def resplit_data(
        self, val_fraction: float, write_csv: bool
    ) -> Optional[Tuple[pd.DataFrame, ...]]:  # pragma: no cover
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)
        """
        pass

    ####################################################################################################################
    # BASE METHODS
    ####################################################################################################################
    def create_directory(self) -> None:  # pragma: no cover
        """
        0: create directory for dataset
        """
        directory_path = join(self.dataset_path, "analyze_data")
        os.makedirs(directory_path, exist_ok=True)

        bash_cmd = (
            f'echo "*" > {env_variable("DIR_DATASETS")}/{self.ner_dataset}/.gitignore'
        )
        try:
            subprocess.run(bash_cmd, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(e)

    def create_ner_tag_mapping_json(self, modify: bool) -> None:  # pragma: no cover
        """
        II: create customized ner_training tag mapping to map tags in original data to tags in formatted data

        Args:
            modify:      [bool], if True: modify tags as specified in method modify_ner_tag_mapping()

        Returns: -
        """
        if modify:
            ner_tag_mapping = self.create_ner_tag_mapping()
        else:
            ner_tag_mapping = dict()

        json_path = join(self.dataset_path, "ner_tag_mapping.json")
        with open(json_path, "w") as f:
            json.dump(ner_tag_mapping, f)

        print(f"> dumped the following dict to {json_path}:")
        print(ner_tag_mapping)

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    def _read_original_file(self, phase: str) -> SENTENCES_ROWS_PRETOKENIZED:
        """
        III: format data

        Args:
            phase: 'train', 'val', 'test'

        Returns:
            sentences_rows: e.g. (-pretokenized-)
                            [
                                [['Inger', 'PER'], ['säger', '0'], .., []],
                                [['Det', '0'], .., []]
                            ]
        """
        self.set_original_file_paths()
        file_path_original = join(self.dataset_path, self.file_name[phase])

        _sentences_rows = list()
        if isfile(file_path_original):
            _sentence = list()
            with open(file_path_original) as f:
                for row in f.readlines():
                    row_list = self._parse_row(row)
                    if len(row_list) > 0:
                        row_list_formatted = self._format_original_file(row_list)
                        if row_list_formatted is not None:
                            _sentence.append(row_list_formatted)
                    else:
                        if len(_sentence):
                            _sentences_rows.append(_sentence)
                        _sentence = list()
            print(f"\n> read {file_path_original}")
        else:  # pragma: no cover
            raise Exception(f"ERROR! could not find file {file_path_original}!")

        return _sentences_rows

    ####################################################################################################################
    # HELPER: WRITE FORMATTED
    ####################################################################################################################
    def _write_formatted_csv(
        self, phase: str, sentences_rows: SENTENCES_ROWS_PRETOKENIZED
    ) -> None:  # pragma: no cover
        """
        III: format data

        Args:
            phase: 'train', 'val', 'test'
            sentences_rows: e.g. (-pretokenized-)
                            [
                                [['Inger', 'PER'], ['säger', '0'], .., []],
                                [['Det', '0'], .., []]
                            ]

        Returns: -
        """
        sentences_rows_formatted = self._format_sentences_rows(sentences_rows)

        df = pd.DataFrame(sentences_rows_formatted)
        file_path = join(self.dataset_path, f"{phase}_formatted.csv")
        df.to_csv(file_path, sep="\t", header=False, index=False)
        print(f"> phase = {phase}: wrote {len(df)} sentences to {file_path}")

    def _write_formatted_jsonl(
        self, phase: str, sentences_rows: SENTENCES_ROWS_UNPRETOKENIZED
    ) -> None:  # pragma: no cover
        """
        save to jsonl file

        Args:
            phase: 'train', 'val', 'test'
            sentences_rows: e.g. (-unpretokenized-)
                            [
                                {
                                    'text': 'Inger säger ..',
                                    'tags': [{'token': 'Inger', 'tag': 'PER', 'char_start': 0, 'char_end': 5}, ..],
                                },
                                {
                                    'text': 'Det ..',
                                    'tags': [{..}, ..]
                                }
                            ]

        Returns: -
        """
        file_path = join(self.dataset_path, f"{phase}_formatted.jsonl")
        with open(file_path, "w") as file:
            for sentence_row in sentences_rows:
                file.write(json.dumps(sentence_row, ensure_ascii=False) + "\n")
        print(
            f"> phase = {phase}: wrote {len(sentences_rows)} sentences to {file_path}"
        )

    def _format_sentences_rows(
        self, sentences_rows: SENTENCES_ROWS_PRETOKENIZED
    ) -> List[Tuple[str, str]]:
        """
        III: format data

        Args:
            sentences_rows: e.g. (-pretokenized-)
                            [
                                [['Inger', 'PER'], ['säger', '0'], .., []],
                                [['Det', '0'], .., []]
                            ]

        Returns:
            sentences_rows_formatted, e.g. (-pretokenized-)
                            [
                                ('PER O', 'Inger säger'),
                                ('O', 'Det'),
                            ]
        """

        # ner tag mapping
        ner_tag_mapping = get_ner_tag_mapping(
            path=join(self.dataset_path, "ner_tag_mapping.json")
        )

        # processing
        sentences_rows_formatted = list()
        for sentence in sentences_rows:
            text_list = list()
            tags_list = list()
            for row in sentence:
                assert (
                    len(row) == 2
                ), f"ERROR! row with length = {len(row)} found (should be 2): {row}"
                text_list.append(row[0])
                tags_list.append(
                    ner_tag_mapping(row[1]) if row[1] != "0" else "O"
                )  # replace zeros by capital O (!)
            sentences_rows_formatted.append((" ".join(tags_list), " ".join(text_list)))

        return sentences_rows_formatted

    @staticmethod
    def _convert_iob1_to_iob2(
        sentences_rows_iob1: SENTENCES_ROWS_PRETOKENIZED,
    ) -> SENTENCES_ROWS_PRETOKENIZED:
        """
        III: format data
        convert tags from IOB1 to IOB2 format

        Args:
            sentences_rows_iob1: e.g. (-pretokenized-)
                            [
                                [['Inger', 'I-PER'], ['säger', '0'], .., []],
                            ]

        Returns:
            sentences_rows_iob2: e.g. (-pretokenized-)
                            [
                                [['Inger', 'B-PER'], ['säger', '0'], .., []],
                            ]
        """
        sentences_rows_iob2 = list()
        for sentence in sentences_rows_iob1:
            sentence_iob2 = list()
            for i, row in enumerate(sentence):
                assert (
                    len(row) == 2
                ), f"ERROR! row = {row} should have length 0 or 2, not {len(row)}"
                current_tag = row[1]

                if (
                    current_tag == "O"
                    or "-" not in current_tag
                    or current_tag.startswith("B-")
                ):
                    sentence_iob2.append(row)
                elif current_tag.startswith("I-"):
                    previous_tag = (
                        sentence[i - 1][1]
                        if (i > 0 and len(sentence[i - 1]) == 2)
                        else None
                    )

                    if previous_tag not in [
                        current_tag,
                        current_tag.replace("I-", "B-"),
                    ]:
                        tag_iob2 = current_tag.replace("I-", "B-")
                        sentence_iob2.append([row[0], tag_iob2])
                    else:
                        sentence_iob2.append(row)

            sentences_rows_iob2.append(sentence_iob2)

        return sentences_rows_iob2

    @staticmethod
    def _shuffle_dataset(_phase: str, _sentences_rows: List[Any]) -> List[Any]:
        """
        III: format data

        Args:
            _phase: "train", "val", "test"
            _sentences_rows: e.g. (-pretokenized-)
                            [
                                [['Inger', 'PER'], ['säger', '0'], .., []],
                                [['Det', '0'], .., []]
                            ]
                            e.g. (-unpretokenized-)
                            [
                                {
                                    'text': 'Inger säger ..',
                                    'tags': [{'token': 'Inger', 'tag': 'PER', 'char_start': 0, 'char_end': 5}, ..],
                                },
                                {
                                    'text': 'Det ..',
                                    'tags': [{..}, ..]
                                }
                            ]

        Returns:
            _sentences_rows_shuffled: e.g. (-pretokenized-)
                            [
                                [['Det', '0'], .., [0],
                                [['Inger', 'PER'], ['säger', '0'], .., []]
                            ]
                            e.g. (-unpretokenized-)
                            [
                                {
                                    'text': 'Det ..',
                                    'tags': [{..}, ..]
                                },
                                {
                                    'text': 'Inger säger ..',
                                    'tags': [{'token': 'Inger', 'tag': 'PER', 'char_start': 0, 'char_end': 5}, ..],
                                }
                            ]
        """
        # change _sentences_rows by shuffling sentences
        random.Random(SEED_SHUFFLE[_phase]).shuffle(_sentences_rows)
        return _sentences_rows

    ####################################################################################################################
    # HELPER: READ FORMATTED
    ####################################################################################################################
    def _read_formatted_csvs(self, phases: List[str]) -> pd.DataFrame:
        """
        IV: resplit data

        Args:
            phases: .. to read formatted csvs from, e.g. ['val', 'test']

        Returns:
            df_phases: contains formatted csvs of phases
        """
        df_phases = [self._read_formatted_csv(phase) for phase in phases]
        return pd.concat(df_phases, ignore_index=True)

    def _read_formatted_csv(self, phase: str):
        """
        IV: resplit data

        Args:
            phase: .. to read formatted df from, e.g. 'val'

        Returns:
            df: formatted df
        """
        formatted_file_path = join(self.dataset_path, f"{phase}_formatted.csv")
        return pd.read_csv(formatted_file_path, sep="\t", header=None)

    ####################################################################################################################
    # HELPER: WRITE FINAL
    ####################################################################################################################
    @staticmethod
    def _split_off_validation_set(
        _df_original: pd.DataFrame, _val_fraction: float
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        IV: resplit data

        Args:
            _df_original:  df before splitting
            _val_fraction: between 0 and 1

        Returns:
            _df_new:       df after splitting, remainder
            _df_val:       df after splitting, validation
        """
        split_index = int(len(_df_original) * (1 - _val_fraction))
        _df_new = _df_original.iloc[:split_index]
        _df_val = _df_original.iloc[split_index:]
        return _df_new, _df_val

    def _write_final_csv(self, phase, df) -> None:  # pragma: no cover
        """
        IV: resplit data

        Args:
            phase: [str], 'train', 'val' or 'test'
            df: [pd DataFrame]

        Returns: -
        """
        file_path = join(self.dataset_path, f"{phase}.csv")
        df.to_csv(file_path, sep="\t", index=False, header=None)



import pytest
from typing import List, Optional

from nerblackbox.modules.utils.util_functions import (
    checkpoint2epoch,
    epoch2checkpoint,
    get_run_name,
    get_run_name_nr,
    compute_mean_and_dmean,
)
from nerblackbox.tests.test_utils import pytest_approx


class TestUtilFunctions:
    @pytest.mark.parametrize(
        "checkpoint_name, epoch",
        [
            ("epoch=2.ckpt", 2),
            ("epoch=35.ckpt", 35),
        ],
    )
    def test_checkpoint2epoch(self, checkpoint_name: str, epoch: int):
        test_epoch = checkpoint2epoch(checkpoint_name)
        assert test_epoch == epoch, f"ERROR! test_epoch = {test_epoch} != {epoch}"

        test_checkpoint_name = epoch2checkpoint(epoch)
        assert (
            test_checkpoint_name == checkpoint_name
        ), f"ERROR! test_checkpoint_name = {test_checkpoint_name} != {checkpoint_name}"

    @pytest.mark.parametrize(
        "run_name, run_nr, run_name_nr",
        [
            ("runA", 3, "runA-3"),
        ],
    )
    def test_get_run_name(self, run_name: str, run_nr: int, run_name_nr: str):
        test_run_name = get_run_name(run_name_nr)
        assert (
            test_run_name == run_name
        ), f"ERROR! test_run_name = {test_run_name} != {run_name}"

        test_run_name_nr = get_run_name_nr(run_name, run_nr)
        assert (
            test_run_name_nr == run_name_nr
        ), f"ERROR! test_run_name_nr = {test_run_name_nr} != {run_name_nr}"

    @pytest.mark.parametrize(
        "list_of_float, " "mean, " "dmean",
        [
            ([], -1.0, None),
            ([1.0], 1.0, None),
            ([1.0, 1.0], 1.0, 0.0),
            ([1.0, 2.0], 1.5, 0.35355),  # = 0.5 / sqrt(2)
            (
                [0.7465069860279442, 0.7784431137724552],
                0.7624750499001998,
                0.0112911262464918,
            ),
        ],
    )
    def test_compute_mean_and_dmean(
        self,
        list_of_float: List[float],
        mean: float,
        dmean: Optional[float],
    ):
        test_mean, test_dmean = compute_mean_and_dmean(list_of_float)
        assert test_mean == pytest_approx(
            mean
        ), f"ERROR! test_mean = {test_mean} != {mean}"
        assert test_dmean == pytest_approx(
            dmean
        ), f"ERROR! test_dmean = {test_dmean} != {dmean}"



import requests
from os.path import join, isfile
from typing import Dict, List, Optional, Tuple
import pandas as pd

from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS,
)


class CoNLL2003Formatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "conll2003"
        ner_tag_list = ["PER", "ORG", "LOC", "MISC"]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]
        """
        url_base = "https://raw.githubusercontent.com/patverga/torch-ner-nlp-from-scratch/master/data/conll2003/"
        targets = ["eng.train", "eng.testa", "eng.testb"]

        for target in targets:
            target_file = join(self.dataset_path, target)

            # fetch tgz from url
            if isfile(target_file):
                if verbose:
                    print(f".. file at {target_file} already exists")
            else:
                url = url_base + target
                myfile = requests.get(url, allow_redirects=True)
                open(target_file, "wb").write(myfile.content)
                if verbose:
                    print(f".. file fetched from {url} and saved at {target_file}")

    def create_ner_tag_mapping(self) -> Dict[str, str]:
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            sentences_rows_iob2: only if write_csv = False
        """
        for phase in ["train", "val", "test"]:
            sentences_rows_iob1 = self._read_original_file(phase)
            if shuffle:
                sentences_rows_iob1 = self._shuffle_dataset(phase, sentences_rows_iob1)

            sentences_rows_iob2 = self._convert_iob1_to_iob2(sentences_rows_iob1)
            if write_csv:  # pragma: no cover
                self._write_formatted_csv(phase, sentences_rows_iob2)
            else:
                return sentences_rows_iob2
        return None

    def set_original_file_paths(self) -> None:
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        self.file_name = {
            "train": "eng.train",
            "val": "eng.testa",
            "test": "eng.testb",
        }

    def _parse_row(self, _row: str) -> List[str]:
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        return _row.strip().split()

    def _format_original_file(self, _row_list: List[str]) -> Optional[List[str]]:
        """
        III: format data

        Args:
            _row_list: e.g. ["test", "PER", "X", "B"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        if not len(_row_list) in [4]:  # pragma: no cover
            raise Exception(f"ERROR! row_list = {_row_list} should consist of 4 parts!")

        if _row_list[0] != "-DOCSTART-":
            _row_list_formatted = [_row_list[0], _row_list[-1]]
            return _row_list_formatted
        else:
            return None

    def resplit_data(
        self, val_fraction: float = 0.0, write_csv: bool = True
    ) -> Optional[Tuple[pd.DataFrame, ...]]:
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            df_train: only if write_csv = False
            df_val:   only if write_csv = False
            df_test:  only if write_csv = False
        """
        # train -> train
        df_train = self._read_formatted_csvs(["train"])

        # val  -> val
        df_val = self._read_formatted_csvs(["val"])

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])

        if write_csv:  # pragma: no cover
            self._write_final_csv("val", df_val)
            self._write_final_csv("train", df_train)
            self._write_final_csv("test", df_test)
            return None
        else:
            return df_train, df_val, df_test



from typing import Dict
from copy import deepcopy
from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
    InputExamples,
)
from nerblackbox.modules.ner_training.annotation_tags.tags import Tags


class InputExamplesUtils:

    ####################################################################################################################
    # 1. from data preprocessing
    ####################################################################################################################
    @classmethod
    def convert_annotation_scheme(
        cls,
        input_examples: Dict[str, InputExamples],
        annotation_scheme_source: str,
        annotation_scheme_target: str,
    ) -> Dict[str, InputExamples]:
        """
        convert input_examples from annotation_scheme_source to annotation_scheme_target

        Args:
            input_examples: [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                      values = [list] of [InputExample]
            annotation_scheme_source: [str], e.g. plain / bio
            annotation_scheme_target: [str], e.g. bio / plain

        Returns:
            input_examples_converted: [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                                values = [list] of [InputExample]
        """
        if annotation_scheme_source == annotation_scheme_target:
            return input_examples
        else:
            input_examples_converted = deepcopy(input_examples)
            for key in input_examples_converted.keys():
                for input_example_converted in input_examples_converted[key]:
                    tags = Tags(input_example_converted.tags.split())
                    input_example_converted.tags = " ".join(
                        tags.convert_scheme(
                            source_scheme=annotation_scheme_source,
                            target_scheme=annotation_scheme_target,
                        )
                    )

        return input_examples_converted



"""Command Line Interface of the nerblackbox package."""

import os
import subprocess
from os.path import join
import click
from nerblackbox.api.store import Store
from nerblackbox.api.dataset import Dataset
from nerblackbox.api.experiment import Experiment
from nerblackbox.api.model import Model
from nerblackbox.modules.experiment_results import ExperimentResults
from nerblackbox.modules.utils.env_variable import env_variable


########################################################################################################################
# CLI
########################################################################################################################
@click.group()
@click.option(
    "--store_dir",
    default="store",
    type=str,
    help="[str] relative path of store directory",
)
@click.option(
    "--modify/--no-modify", default=False, help="[bool] if flag=set_up_dataset"
)
@click.option(
    "--val_fraction", default=None, type=float, help="[float] if flag=set_up_dataset"
)
@click.option(
    "--verbose/--no-verbose", default=False, help="[bool] if flag=set_up_dataset"
)
@click.option("--run_name", default=None, type=str, help="[str] if flag=run_experiment")
@click.option("--device", default=None, type=str, help="[str] if flag=run_experiment")
@click.option("--fp16/--no-fp16", default=False, help="[bool] if flag=run_experiment")
@click.option("--results/--no-results", default=False, help="[bool] if flag=clear_data")
@click.pass_context
def nerbb(ctx, **kwargs_optional):
    ctx.ensure_object(dict)

    # kwargs
    kwargs = {k: v for k, v in kwargs_optional.items() if v is not None}

    # environ
    data_dir = kwargs.pop("store_dir")
    if len(data_dir):
        Store.set_path(data_dir)

    # context
    ctx.obj = kwargs


########################################################################################################################
# COMMANDS
########################################################################################################################
@nerbb.command(name="analyze_data")
@click.argument("dataset_name")
def analyze_data(dataset_name: str):
    """analyze a dataset."""
    dataset = Dataset(dataset_name)
    dataset.overview()


@click.pass_context
@nerbb.command(name="clear_data")
def clear_data(ctx):
    """clear data (checkpoints and optionally results)."""
    context = {k: v for k, v in ctx.obj.items() if k in ["results"]}
    Store.clear_data(**context)


@nerbb.command(name="create")
def create():
    """
    initialize the data_dir directory.
    needs to be called exactly once before any other CLI/API commands of the package are executed.
    """
    Store.create()


@nerbb.command(name="show_experiments")
def show_experiments():
    """get overview on experiments."""
    Store.show_experiments()


@nerbb.command(name="get_experiment_results")
@click.argument("experiment_name")
def get_experiment_results(experiment_name: str):
    """get results for a single experiment."""
    experiment = Experiment(experiment_name)
    assert isinstance(
        experiment.results, ExperimentResults
    ), f"ERROR! experiment.results is not an instance of ExperimentResults."
    for attribute in ["best_single_run"]:
        assert hasattr(
            experiment.results, attribute
        ), f"ERROR! experiment.results does not have attribute = {attribute}"

    for average in [False, True]:
        score = experiment.get_result(
            metric="f1", level="entity", label="micro", phase="test", average=average
        )
        print(f"score (average={average}) = {score}")
        assert isinstance(
            score, str
        ), f"ERROR! experiment.get_result() did not return a str for average = {average}."


@nerbb.command(name="mlflow")
def mlflow():
    """show detailed experiment results in mlflow (port = 5000)."""
    cd_dir = f'{join(env_variable("DATA_DIR"), "results")}'
    subprocess.run(f"cd {cd_dir}; mlflow ui", shell=True)


@nerbb.command(name="predict")
@click.argument("experiment_name")
@click.argument("text_input")
def predict(experiment_name: str, text_input: str):
    """predict labels for text_input using the best model of a single experiment."""
    model = Model.from_experiment(experiment_name)
    assert isinstance(
        model, Model
    ), f"ERROR! model from experiment {experiment_name} could not be loaded."
    predictions = model.predict(text_input)
    print(predictions)


@nerbb.command(name="predict_proba")
@click.argument("experiment_name")
@click.argument("text_input")
def predict_proba(experiment_name: str, text_input: str):
    """predict label probabilities for text_input using the best model of a single experiment."""
    model = Model.from_experiment(experiment_name)
    assert isinstance(
        model, Model
    ), f"ERROR! model from experiment {experiment_name} could not be loaded."
    predictions = model.predict_proba(text_input)
    print(predictions)


@nerbb.command(name="run_experiment")
@click.pass_context
@click.argument("experiment_name")
def run_experiment(ctx, experiment_name: str):
    """run a single experiment."""
    context = {k: v for k, v in ctx.obj.items() if k in ["run_name", "device", "fp16"]}
    experiment = Experiment(experiment_name, from_config=True, **context)
    experiment.run()


@nerbb.command(name="set_up_dataset")
@click.pass_context
@click.argument("dataset_and_subset_name")
def set_up_dataset(ctx, dataset_and_subset_name: str):
    """set up a dataset using the associated Formatter class."""
    split_list = dataset_and_subset_name.split()
    dataset = Dataset(
        dataset_name=split_list[0],
        dataset_subset_name=split_list[-1] if len(split_list) > 1 else "",
    )
    context = {
        k: v for k, v in ctx.obj.items() if k in ["modify", "val_fraction", "verbose"]
    }
    dataset.set_up(**context)


@nerbb.command(name="show_experiment_config")
@click.argument("experiment_name")
def show_experiment_config(experiment_name: str):
    """show a single experiment configuration in detail."""
    experiment = Experiment(experiment_name)
    experiment.show_config()


@nerbb.command(name="tensorboard")
def tensorboard():
    """show detailed experiment results in tensorboard. (port = 6006)."""
    cd_dir = f'{join(env_variable("DATA_DIR"), "results")}'
    subprocess.run(
        f"cd {cd_dir}; tensorboard --logdir tensorboard --reload_multifile=true",
        shell=True,
    )



import pytest
from typing import List, Tuple

import pandas as pd
from nerblackbox.modules.datasets.formatter.auto_formatter import AutoFormatter
from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS_PRETOKENIZED,
)
from nerblackbox.modules.datasets.formatter.conll2003_formatter import (
    CoNLL2003Formatter,
)
from nerblackbox.modules.datasets.formatter.swe_nerc_formatter import SweNercFormatter
from nerblackbox.modules.datasets.formatter.swedish_ner_corpus_formatter import (
    SwedishNerCorpusFormatter,
)
from nerblackbox.modules.datasets.formatter.sic_formatter import SICFormatter
from nerblackbox.modules.datasets.formatter.suc_formatter import SUCFormatter
from nerblackbox.modules.datasets.formatter.sucx_formatter import SUCXFormatter
from nerblackbox.modules.datasets.formatter.huggingface_datasets_formatter import (
    HuggingfaceDatasetsFormatter,
)

from pkg_resources import resource_filename
import os
from os.path import abspath, dirname, join

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR


class TestAutoFormatter:
    @pytest.mark.parametrize(
        "ner_dataset, ner_dataset_subset, error",
        [
            ("swedish_ner_corpus", "", False),
            ("conll2003", "", False),
            ("sic", "", False),
            ("suc", "", False),
            ("suc", "xyz", False),  # ner_dataset_subset is not used
            ("sucx", "original_cased", False),
            ("sucx", "", True),
            ("swe_nerc", "", False),
            ("xyz", "", True),
            ("ehealth_kd", "", False),
            ("sent_comp", "", True),
        ],
    )
    def test_for_dataset(self, ner_dataset: str, ner_dataset_subset: str, error: bool):
        if error:
            with pytest.raises(Exception):
                _ = AutoFormatter.for_dataset(
                    ner_dataset=ner_dataset, ner_dataset_subset=ner_dataset_subset
                )
        else:
            auto_formatter = AutoFormatter.for_dataset(
                ner_dataset=ner_dataset, ner_dataset_subset=ner_dataset_subset
            )
            assert isinstance(
                auto_formatter, BaseFormatter
            ), f"ERROR! type(auto_formatter) = {type(auto_formatter)} != BaseFormatter"
            assert (
                auto_formatter.ner_dataset == ner_dataset
            ), f"ERROR! auto_formatter.ner_dataset = {auto_formatter.ner_dataset} != {ner_dataset}"


class TestBaseFormatter:

    base_formatter = SwedishNerCorpusFormatter()
    base_formatter.dataset_path = resource_filename(
        "nerblackbox", f"tests/test_data/formatted_data"
    )

    @pytest.mark.parametrize(
        "phases, df_formatted",
        [
            (
                ["train"],
                pd.DataFrame(
                    data=[
                        ["O O", "Mening 1"],
                        ["PER O", "Mening 2"],
                        ["O PER", "Mening 3"],
                        ["O PER", "Mening 4"],
                    ]
                ),
            ),
            (["val"], pd.DataFrame(data=[["O O", "Mening 5"], ["PER O", "Mening 6"]])),
            (["test"], pd.DataFrame(data=[["O O", "Mening 7"], ["PER O", "Mening 8"]])),
            (
                ["val", "test"],
                pd.DataFrame(
                    data=[
                        ["O O", "Mening 5"],
                        ["PER O", "Mening 6"],
                        ["O O", "Mening 7"],
                        ["PER O", "Mening 8"],
                    ]
                ),
            ),
            (
                ["test", "val"],
                pd.DataFrame(
                    data=[
                        ["O O", "Mening 7"],
                        ["PER O", "Mening 8"],
                        ["O O", "Mening 5"],
                        ["PER O", "Mening 6"],
                    ]
                ),
            ),
        ],
    )
    def test_read_formatted_csvs(self, phases: List[str], df_formatted: pd.DataFrame):
        test_df_formatted = self.base_formatter._read_formatted_csvs(phases)
        pd.testing.assert_frame_equal(
            test_df_formatted, df_formatted
        ), f"ERROR! test_read_formatted did not pass test for phases = {phases}"

    @pytest.mark.parametrize(
        "df_original, val_fraction, df_new, df_val",
        [
            (
                pd.DataFrame(data=[[1, 2], [3, 4]], columns=["A", "B"], index=[0, 1]),
                0.5,
                pd.DataFrame(data=[[1, 2]], columns=["A", "B"], index=[0]),
                pd.DataFrame(data=[[3, 4]], columns=["A", "B"], index=[1]),
            ),
            (
                pd.DataFrame(
                    data=[[1, 2], [3, 4], [5, 6], [7, 8]],
                    columns=["A", "B"],
                    index=["a", "b", "c", "d"],
                ),
                0.25,
                pd.DataFrame(
                    data=[[1, 2], [3, 4], [5, 6]],
                    columns=["A", "B"],
                    index=["a", "b", "c"],
                ),
                pd.DataFrame(data=[[7, 8]], columns=["A", "B"], index=["d"]),
            ),
        ],
    )
    def test_split_off_validation_set(
        self,
        df_original: pd.DataFrame,
        val_fraction: float,
        df_new: pd.DataFrame,
        df_val: pd.DataFrame,
    ):
        test_df_new, test_df_val = self.base_formatter._split_off_validation_set(
            df_original, val_fraction
        )
        pd.testing.assert_frame_equal(
            test_df_new, df_new
        ), f"ERROR! test_split_off_validation_set / df_new did not pass test"
        pd.testing.assert_frame_equal(
            test_df_val, df_val
        ), f"ERROR! test_split_off_validation_set / df_val did not pass test"

    @pytest.mark.parametrize(
        "sentences_rows, sentences_rows_formatted",
        [
            (
                [
                    [["Inger", "B-PER"], ["Nilsson", "I-PER"], ["Frida", "B-PER"]],
                    [["är", "O"], ["Inger", "B-PER"], ["och", "O"]],
                ],
                [
                    ("B-PER I-PER B-PER", "Inger Nilsson Frida"),
                    ("O B-PER O", "är Inger och"),
                ],
            )
        ],
    )
    def test_format_sentences_rows(
        self,
        sentences_rows: SENTENCES_ROWS_PRETOKENIZED,
        sentences_rows_formatted: List[Tuple[str, str]],
    ):
        test_sentences_rows_formatted = self.base_formatter._format_sentences_rows(
            sentences_rows
        )
        assert (
            test_sentences_rows_formatted == sentences_rows_formatted
        ), f"ERROR! test_sentences_rows_formatted = {test_sentences_rows_formatted} != {sentences_rows_formatted}"

    @pytest.mark.parametrize(
        "sentences_rows_iob1, sentences_rows_iob2",
        [
            (
                [[["Inger", "I-PER"], ["säger", "0"]]],
                [[["Inger", "B-PER"], ["säger", "0"]]],
            ),
            (
                [[["Inger", "I-PER"], ["Nilsson", "I-PER"], ["Frida", "B-PER"]]],
                [[["Inger", "B-PER"], ["Nilsson", "I-PER"], ["Frida", "B-PER"]]],
            ),
            (
                [[["är", "O"], ["Inger", "I-PER"], ["och", "O"]]],
                [[["är", "O"], ["Inger", "B-PER"], ["och", "O"]]],
            ),
        ],
    )
    def test_convert_iob1_to_iob2(
        self,
        sentences_rows_iob1: SENTENCES_ROWS_PRETOKENIZED,
        sentences_rows_iob2: SENTENCES_ROWS_PRETOKENIZED,
    ):
        test_sentences_rows_iob2 = self.base_formatter._convert_iob1_to_iob2(
            sentences_rows_iob1
        )
        assert (
            test_sentences_rows_iob2 == sentences_rows_iob2
        ), f"ERROR! test_sentences_rows_iob2 = {test_sentences_rows_iob2} != {sentences_rows_iob2}"

    @pytest.mark.parametrize(
        "phase, sentences_rows, sentences_rows_shuffled",
        [
            (
                "train",
                [
                    [["Inger", "B-PER"], ["säger", "0"]],
                    [["Inger", "B-PER"], ["Nilsson", "I-PER"], ["Frida", "B-PER"]],
                    [["är", "O"], ["Inger", "B-PER"], ["och", "O"]],
                ],
                [
                    [["är", "O"], ["Inger", "B-PER"], ["och", "O"]],
                    [["Inger", "B-PER"], ["Nilsson", "I-PER"], ["Frida", "B-PER"]],
                    [["Inger", "B-PER"], ["säger", "0"]],
                ],
            ),
        ],
    )
    def test_shuffle_dataset(
        self,
        phase: str,
        sentences_rows: SENTENCES_ROWS_PRETOKENIZED,
        sentences_rows_shuffled: SENTENCES_ROWS_PRETOKENIZED,
    ):
        test_shuffled_sentences_rows = self.base_formatter._shuffle_dataset(
            phase, sentences_rows
        )
        assert test_shuffled_sentences_rows == sentences_rows_shuffled, (
            f"ERROR! test_shuffled_sentences_rows = {test_shuffled_sentences_rows} "
            f"!= {sentences_rows_shuffled} = sentences_rows_shuffled"
        )


class TestAllFormatters:

    formatters = {
        "conll2003": CoNLL2003Formatter(),
        "swe_nerc": SweNercFormatter(),
        "swedish_ner_corpus": SwedishNerCorpusFormatter(),
        "sic": SICFormatter(),
        "suc": SUCFormatter(),
        "sucx": SUCXFormatter("original_cased"),
        "ehealth_kd": HuggingfaceDatasetsFormatter("ehealth_kd"),
    }

    @pytest.mark.parametrize(
        "ner_dataset",
        [
            "conll2003",
            "swe_nerc",
            "swedish_ner_corpus",
            "sic",
            "suc",
            "sucx",
            "ehealth_kd",
        ],
    )
    def test_create_ner_tag_mapping(self, ner_dataset: str):
        ner_tag_mapping = self.formatters[ner_dataset].create_ner_tag_mapping()
        assert isinstance(
            ner_tag_mapping, dict
        ), f"ERROR! ner_tag_mapping = {ner_tag_mapping} for ner_dataset = {ner_dataset} is not a dict!"
